import os
import torch
import logging
import typing as _typing
import torch.nn.functional
import torch.utils.data

import autogl.data
from .. import register_trainer
from ..base import BaseNodeClassificationTrainer, EarlyStopping, Evaluation
from ..evaluation import get_feval, EvaluatorUtility, Logloss, MicroF1
from ..sampling.sampler.target_dependant_sampler import TargetDependantSampledData
from ..sampling.sampler.neighbor_sampler import NeighborSampler
from ..sampling.sampler.graphsaint_sampler import *
from ..sampling.sampler.layer_dependent_importance_sampler import (
    LayerDependentImportanceSampler,
)
from ...model import BaseAutoModel
from ...model.base import ClassificationSupportedSequentialModel

LOGGER: logging.Logger = logging.getLogger("Node classification sampling trainer")


class _DeterministicNeighborSamplerStore:
    def __init__(self):
        self.__neighbor_sampler_mapping: _typing.List[
            _typing.Tuple[torch.LongTensor, NeighborSampler]
        ] = []

    @classmethod
    def __is_target_node_indexes_equal(
        cls, a: torch.LongTensor, b: torch.LongTensor
    ) -> bool:
        if not a.dtype == b.dtype == torch.int64:
            return False
        if a.size() != b.size():
            return False
        return torch.where(a != b)[0].size(0) == 0

    def __setitem__(
        self, target_nodes: torch.Tensor, neighbor_sampler: NeighborSampler
    ):
        target_nodes: _typing.Any = target_nodes.cpu()
        if type(target_nodes) != torch.Tensor or target_nodes.dtype != torch.int64:
            raise TypeError
        if type(neighbor_sampler) != NeighborSampler:
            raise TypeError
        for i in range(len(self.__neighbor_sampler_mapping)):
            if self.__is_target_node_indexes_equal(
                target_nodes, self.__neighbor_sampler_mapping[i][0]
            ):
                self.__neighbor_sampler_mapping[i] = (target_nodes, neighbor_sampler)
                return
        self.__neighbor_sampler_mapping.append((target_nodes, neighbor_sampler))

    def __getitem__(
        self, target_nodes: torch.Tensor
    ) -> _typing.Optional[NeighborSampler]:
        target_nodes: _typing.Any = target_nodes.cpu()
        if type(target_nodes) != torch.Tensor or target_nodes.dtype != torch.int64:
            raise TypeError
        for (
            __current_target_nodes,
            __neighbor_sampler,
        ) in self.__neighbor_sampler_mapping:
            if self.__is_target_node_indexes_equal(
                target_nodes, __current_target_nodes
            ):
                return __neighbor_sampler
        return None


@register_trainer("NodeClassificationGraphSAINTTrainer")
class NodeClassificationGraphSAINTTrainer(BaseNodeClassificationTrainer):
    """
    The node classification trainer utilizing GraphSAINT technique.

    Parameters
    ------------
    model: ``BaseModel`` or ``str``
        The name or class of model adopted
    num_features: ``int``
        number of features for each node provided by dataset
    num_classes: ``int``
        number of classes to classify
    optimizer: ``Optimizer`` of ``str``
        The (name of) optimizer used to train and predict.
    lr: ``float``
        The learning rate of link prediction task.
    max_epoch: ``int``
        The max number of epochs in training.
    early_stopping_round: ``int``
        The round of early stop.
    weight_decay: ``float``
        The weight decay argument for optimizer
    device: ``torch.device`` or ``str``
        The device where model will be running on.
    init: ``bool``
        If True(False), the model will (not) be initialized.
    feval: ``str``.
        The evaluation method adopted in this function.
    """

    def __init__(
        self,
        model: _typing.Union[BaseAutoModel, str],
        num_features: int,
        num_classes: int,
        optimizer: _typing.Union[_typing.Type[torch.optim.Optimizer], str, None] = ...,
        lr: float = 1e-4,
        max_epoch: int = 100,
        early_stopping_round: int = 100,
        weight_decay: float = 1e-4,
        device: _typing.Optional[torch.device] = None,
        init: bool = True,
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (MicroF1,),
        loss: str = "nll_loss",
        lr_scheduler_type: _typing.Optional[str] = None,
        **kwargs,
    ):
        if isinstance(optimizer, type) and issubclass(optimizer, torch.optim.Optimizer):
            self._optimizer_class: _typing.Type[torch.optim.Optimizer] = optimizer
        elif type(optimizer) == str:
            if optimizer.lower() == "adam":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.Adam
            elif optimizer.lower() == "adam" + "w":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.AdamW
            elif optimizer.lower() == "sgd":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.SGD
            else:
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.Adam
        else:
            self._optimizer_class: _typing.Type[
                torch.optim.Optimizer
            ] = torch.optim.Adam
        self._learning_rate: float = lr if lr > 0 else 1e-4
        self._lr_scheduler_type: _typing.Optional[str] = lr_scheduler_type
        self._max_epoch: int = max_epoch if max_epoch > 0 else 1e2
        self._weight_decay: float = weight_decay if weight_decay > 0 else 1e-4
        self._early_stopping = EarlyStopping(
            patience=early_stopping_round if early_stopping_round > 0 else 1e2,
            verbose=False,
        )
        """ Assign an empty initial hyper parameter space """
        self._hyper_parameter_space: _typing.Sequence[
            _typing.Dict[str, _typing.Any]
        ] = []

        self._valid_result: torch.Tensor = torch.zeros(0)
        self._valid_result_prob: torch.Tensor = torch.zeros(0)
        self._valid_score: _typing.Sequence[float] = ()

        """ Set GraphSAINT hyper-parameters """
        " Set sampler_type "
        sampler_type: str = kwargs.get("sampler_type", "edge")
        if type(sampler_type) != str:
            raise TypeError
        else:
            sampler_type: str = sampler_type.strip().lower()
        if sampler_type not in ("node", "edge", "rw"):
            sampler_type: str = "edge"  # default to edge sampler
        self.__sampler_type: str = sampler_type

        " Set num_graphs_per_epoch "
        num_graphs_per_epoch: int = kwargs.get("num_graphs_per_epoch", 50)
        if type(num_graphs_per_epoch) != int:
            raise TypeError
        elif not num_graphs_per_epoch > 0:
            num_graphs_per_epoch = 50
        self.__num_graphs_per_epoch: int = num_graphs_per_epoch

        " Set sampled_budget "
        sampled_budget: int = kwargs.get("sampled_budget", 1e4)
        # todo: This is a version caused by current unreasonable initialization process
        # todo: Refactor the framework for trainer to fix in future version
        # if type(sampled_budget) != int:
        #     raise TypeError
        # if not sampled_budget > 0:
        #     raise ValueError
        self.__sampled_budget: int = sampled_budget

        " Set walk_length "
        walk_length: int = kwargs.get("walk_length", 2)
        if type(walk_length) != int:
            raise TypeError
        if not walk_length > 0:
            raise ValueError
        self.__walk_length: int = walk_length

        " Set sample_coverage_factor "
        sample_coverage_factor: int = kwargs.get("sample_coverage_factor", 50)
        if type(sample_coverage_factor) != int:
            raise TypeError
        elif not sample_coverage_factor > 0:
            sample_coverage_factor = 50
        self.__sample_coverage_factor: int = sample_coverage_factor

        """ Set num_workers """

        def _cpu_count() -> int:
            __cpu_count: _typing.Optional[int] = os.cpu_count()
            return __cpu_count if __cpu_count else 0

        # self.__training_sampler_num_workers: int = kwargs.get(
        #     "training_sampler_num_workers", _cpu_count()
        # )

        # if not 0 <= self.__training_sampler_num_workers <= _cpu_count():
        #     self.__training_sampler_num_workers: int = _cpu_count()

        # force to be 0 to be compactible with current pyg solution.
        self.__training_sampler_num_workers: int = 0

        super(NodeClassificationGraphSAINTTrainer, self).__init__(
            model, None, num_features, num_classes, device, feval, loss
        )
        self.__is_initialized: bool = False
        if init:
            self.initialize()

    def initialize(self):
        if self.__is_initialized:
            return self
        self.model.initialize()
        self.__is_initialized = True
        return self

    def to(self, device: torch.device):
        self.device = device
        if self.model is not None:
            self.model.to(self.device)

    def get_model(self):
        return self.model

    def __train_only(self, integral_data):
        """
        The function of training on the given dataset and mask.
        :param integral_data: data of a specific graph
        :return: None
        """
        optimizer: torch.optim.Optimizer = self._optimizer_class(
            self.model.model.parameters(),
            lr=self._learning_rate,
            weight_decay=self._weight_decay,
        )
        if type(self._lr_scheduler_type) == str:
            if self._lr_scheduler_type.lower() == "step" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.StepLR = (
                    torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.1)
                )
            elif self._lr_scheduler_type.lower() == "multi" + "step" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.MultiStepLR = (
                    torch.optim.lr_scheduler.MultiStepLR(
                        optimizer, milestones=[30, 80], gamma=0.1
                    )
                )
            elif self._lr_scheduler_type.lower() == "exponential" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.ExponentialLR = (
                    torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.1)
                )
            elif self._lr_scheduler_type.lower() == "ReduceLROnPlateau".lower():
                lr_scheduler: torch.optim.lr_scheduler.ReduceLROnPlateau = (
                    torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, "min")
                )
            else:
                lr_scheduler: torch.optim.lr_scheduler.LambdaLR = (
                    torch.optim.lr_scheduler.LambdaLR(optimizer, lambda _: 1.0)
                )
        else:
            lr_scheduler: torch.optim.lr_scheduler.LambdaLR = (
                torch.optim.lr_scheduler.LambdaLR(optimizer, lambda _: 1.0)
            )

        setattr(
            integral_data,
            "edge_weight",
            self.__compute_normalized_edge_weight(getattr(integral_data, "edge_index")),
        )
        " Generate Sampler "
        if self.__sampler_type.lower() == "edge":
            _sampler: torch_geometric.data.GraphSAINTEdgeSampler = (
                GraphSAINTSamplerFactory.create_edge_sampler(
                    integral_data,
                    self.__num_graphs_per_epoch,
                    self.__sampled_budget,
                    self.__sample_coverage_factor,
                    num_workers=self.__training_sampler_num_workers,
                )
            )
        elif self.__sampler_type.lower() == "node":
            _sampler: torch_geometric.data.GraphSAINTNodeSampler = (
                GraphSAINTSamplerFactory.create_node_sampler(
                    integral_data,
                    self.__num_graphs_per_epoch,
                    self.__sampled_budget,
                    self.__sample_coverage_factor,
                    num_workers=self.__training_sampler_num_workers,
                )
            )
        elif self.__sampler_type.lower() == "rw":
            _sampler: torch_geometric.data.GraphSAINTRandomWalkSampler = (
                GraphSAINTSamplerFactory.create_random_walk_sampler(
                    integral_data,
                    self.__num_graphs_per_epoch,
                    self.__sampled_budget,
                    self.__walk_length,
                    self.__sample_coverage_factor,
                    num_workers=self.__training_sampler_num_workers,
                )
            )
        else:
            _sampler: torch_geometric.data.GraphSAINTRandomWalkSampler = (
                GraphSAINTSamplerFactory.create_random_walk_sampler(
                    integral_data,
                    self.__num_graphs_per_epoch,
                    self.__sampled_budget,
                    self.__walk_length,
                    self.__sample_coverage_factor,
                    num_workers=self.__training_sampler_num_workers,
                )
            )

        for current_epoch in range(self._max_epoch):
            self.model.model.train()
            optimizer.zero_grad()
            """ epoch start """
            for sampled_data in _sampler:
                sampled_data = sampled_data.to(self.device)
                setattr(
                    sampled_data,
                    "edge_weight",
                    getattr(sampled_data, "edge_norm")
                    * getattr(sampled_data, "edge_weight"),
                )
                optimizer.zero_grad()
                if isinstance(self.model.model, ClassificationSupportedSequentialModel):
                    prediction: torch.Tensor = self.model.model.cls_forward(
                        sampled_data
                    )
                else:
                    prediction: torch.Tensor = self.model.model(sampled_data)
                if not hasattr(torch.nn.functional, self.loss):
                    raise TypeError(f"PyTorch does not support loss type {self.loss}")
                loss_function = getattr(torch.nn.functional, self.loss)
                loss_value: torch.Tensor = loss_function(
                    prediction, getattr(sampled_data, "y"), reduction="none"
                )
                loss_value = (loss_value * getattr(sampled_data, "node_norm"))[
                    sampled_data.train_mask
                ].sum()
                loss_value.backward()
                optimizer.step()

            lr_scheduler.step()
            if (
                hasattr(integral_data, "val_mask")
                and getattr(integral_data, "val_mask") is not None
                and type(getattr(integral_data, "val_mask")) == torch.Tensor
            ):
                validation_results: _typing.Sequence[float] = self.evaluate(
                    (integral_data,), "val", [self.feval[0]]
                )
                if self.feval[0].is_higher_better():
                    validation_loss: float = -validation_results[0]
                else:
                    validation_loss: float = validation_results[0]
                self._early_stopping(validation_loss, self.model.model)
                if self._early_stopping.early_stop:
                    LOGGER.debug("Early stopping at %d", current_epoch)
                    break
        if (
            hasattr(integral_data, "val_mask")
            and getattr(integral_data, "val_mask") is not None
            and type(getattr(integral_data, "val_mask")) == torch.Tensor
        ):
            self._early_stopping.load_checkpoint(self.model.model)

    def __predict_only(
        self,
        integral_data,
        mask_or_target_nodes_indexes: _typing.Union[torch.BoolTensor, torch.LongTensor],
    ) -> torch.Tensor:
        """
        The function of predicting on the given data.
        :param integral_data: data of a specific graph
        :param mask_or_target_nodes_indexes: ...
        :return: the result of prediction on the given dataset
        """
        import copy

        integral_data = copy.copy(integral_data)
        self.model.model.eval()
        setattr(
            integral_data,
            "edge_weight",
            self.__compute_normalized_edge_weight(getattr(integral_data, "edge_index")),
        )
        integral_data = integral_data.to(self.device)
        with torch.no_grad():
            if isinstance(self.model.model, ClassificationSupportedSequentialModel):
                prediction: torch.Tensor = self.model.model.cls_forward(integral_data)
            else:
                prediction: torch.Tensor = self.model.model(integral_data)
        return prediction[mask_or_target_nodes_indexes]

    def predict_proba(
        self, dataset, mask: _typing.Optional[str] = None, in_log_format: bool = False
    ):
        """
        The function of predicting the probability on the given dataset.

        Parameters
        ----------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.
        in_log_format: ``bool``.
            If True(False), the probability will (not) be log format.

        Returns
        -------
        The prediction result.
        """
        data = dataset[0].to(torch.device("cpu"))
        if mask is not None and type(mask) == str:
            if mask.lower() == "train":
                _mask: torch.BoolTensor = data.train_mask
            elif mask.lower() == "test":
                _mask: torch.BoolTensor = data.test_mask
            elif mask.lower() == "val":
                _mask: torch.BoolTensor = data.val_mask
            else:
                _mask: torch.BoolTensor = data.test_mask
        else:
            _mask: torch.BoolTensor = data.test_mask
        result = self.__predict_only(data, _mask)
        return result if in_log_format else torch.exp(result)

    def predict(self, dataset, mask: _typing.Optional[str] = None) -> torch.Tensor:
        """
        The function of predicting on the given dataset.

        Parameters
        ----------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.

        Returns
        -------
        The prediction result of ``predict_proba``.
        """
        return self.predict_proba(dataset, mask, in_log_format=True).max(1)[1]

    def evaluate(
        self,
        dataset,
        mask: _typing.Optional[str] = None,
        feval: _typing.Union[
            None, _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = None,
    ) -> _typing.Sequence[float]:
        """
        The function of training on the given dataset and keeping valid result.

        Parameters
        ------------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.
        feval: ``str``.
            The evaluation method adopted in this function.

        Returns
        -------
        result: The evaluation result on the given dataset.
        """
        data = dataset[0]
        data = data.to(self.device)
        if feval is None:
            _feval: _typing.Sequence[_typing.Type[Evaluation]] = self.feval
        else:
            _feval: _typing.Sequence[_typing.Type[Evaluation]] = get_feval(list(feval))
        if mask is not None and type(mask) == str:
            if mask.lower() == "train":
                _mask: torch.Tensor = data.train_mask
            elif mask.lower() == "test":
                _mask: torch.Tensor = data.test_mask
            elif mask.lower() == "val":
                _mask: torch.Tensor = data.val_mask
            else:
                _mask: torch.Tensor = data.test_mask
        else:
            _mask: torch.Tensor = data.test_mask
        prediction_probability: torch.Tensor = self.predict_proba(dataset, mask)
        y_ground_truth: torch.Tensor = data.y[_mask]

        return [
            f.evaluate(
                prediction_probability.cpu().numpy(),
                y_ground_truth.cpu().numpy(),
            )
            for f in _feval
        ]

    @classmethod
    def __compute_normalized_edge_weight(
        cls,
        edge_index: torch.LongTensor,
        original_edge_weight: _typing.Optional[torch.Tensor] = ...,
    ) -> torch.Tensor:
        if type(edge_index) != torch.Tensor:
            raise TypeError
        if original_edge_weight in (None, Ellipsis, ...):
            original_edge_weight: torch.Tensor = torch.ones(edge_index.size(1))
        elif type(original_edge_weight) != torch.Tensor:
            raise TypeError
        elif original_edge_weight.numel() != edge_index.size(1):
            raise ValueError
        elif original_edge_weight.size() != (edge_index.size(1),):
            original_edge_weight = original_edge_weight.resize(edge_index.size(1))

        __out_degree: torch.Tensor = torch_geometric.utils.degree(edge_index[0])
        __in_degree: torch.Tensor = torch_geometric.utils.degree(edge_index[1])
        temp_tensor: torch.Tensor = torch.stack(
            [__out_degree[edge_index[0]], __in_degree[edge_index[1]]]
        )
        temp_tensor: torch.Tensor = torch.pow(temp_tensor, -0.5)
        temp_tensor[torch.isinf(temp_tensor)] = 0.0
        return original_edge_weight * temp_tensor[0] * temp_tensor[1]

    def train(self, dataset, keep_valid_result: bool = True):
        """
        The function of training on the given dataset and keeping valid result.

        Parameters
        ------------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        keep_valid_result: ``bool``
            If True(False), save the validation result after training.

        Returns
        --------
        None
        """
        import gc

        gc.collect()
        data = dataset[0].to(torch.device("cpu"))
        self.__train_only(data)
        if keep_valid_result:
            prediction: torch.Tensor = self.__predict_only(data, data.val_mask)
            self._valid_result: torch.Tensor = prediction.max(1)[1]
            self._valid_result_prob: torch.Tensor = prediction
            self._valid_score: _typing.Sequence[float] = self.evaluate(dataset, "val")

    def get_valid_predict(self) -> torch.Tensor:
        return self._valid_result

    def get_valid_predict_proba(self) -> torch.Tensor:
        return self._valid_result_prob

    def get_valid_score(
        self, return_major: bool = True
    ) -> _typing.Union[
        _typing.Tuple[float, bool],
        _typing.Tuple[_typing.Sequence[float], _typing.Sequence[bool]],
    ]:
        """
        The function of getting the valid score.

        Parameters
        ----------
        return_major: ``bool``.
            If True, the return only consists of the major result.
            If False, the return consists of the all results.

        Returns
        -------
        result: The valid score.
        """
        if return_major:
            return self._valid_score[0], self.feval[0].is_higher_better()
        else:
            return self._valid_score, [f.is_higher_better() for f in self.feval]

    def __repr__(self) -> str:
        import yaml

        __repr: dict = {
            "trainer_name": self.__class__.__name__,
            "learning_rate": self._learning_rate,
            "model": repr(self.model),
            "max_epoch": self._max_epoch,
            "early_stopping_round": self._early_stopping.patience,
            "sampler_type": self.__sampler_type,
            "sampled_budget": self.__sampled_budget,
        }
        if self.__sampler_type == "rw":
            __repr.update({"walk_length": self.__walk_length})
        return yaml.dump(__repr)

    def duplicate_from_hyper_parameter(
        self,
        hp: _typing.Dict[str, _typing.Any],
        model: _typing.Optional[BaseAutoModel] = None,
    ) -> "NodeClassificationGraphSAINTTrainer":
        """
        The function of duplicating a new instance from the given hyper-parameter.

        Parameters
        ------------
        hp: ``dict``.
            The hyper-parameter settings for the new instance.
        model: ``BaseModel``
            The name or class of model adopted

        Returns
        --------
        instance: ``NodeClassificationGraphSAINTTrainer``
            A new instance of trainer.
        """
        if model is None or not isinstance(model, BaseAutoModel):
            model: BaseAutoModel = self.model
        model = model.from_hyper_parameter(hp["encoder"])
        return NodeClassificationGraphSAINTTrainer(
            model,
            self.num_features,
            self.num_classes,
            self._optimizer_class,
            device=self.device,
            init=True,
            feval=self.feval,
            loss=self.loss,
            lr_scheduler_type=self._lr_scheduler_type,
            **hp["trainer"],
        )


@register_trainer("NodeClassificationLayerDependentImportanceSamplingTrainer")
class NodeClassificationLayerDependentImportanceSamplingTrainer(
    BaseNodeClassificationTrainer
):
    """
    The node classification trainer utilizing Layer dependent importance sampling technique.

    Parameters
    ------------
    model: ``BaseModel`` or ``str``
        The name or class of model adopted
    num_features: ``int``
        number of features for each node provided by dataset
    num_classes: ``int``
        number of classes to classify
    optimizer: ``Optimizer`` of ``str``
        The (name of) optimizer used to train and predict.
    lr: ``float``
        The learning rate of link prediction task.
    max_epoch: ``int``
        The max number of epochs in training.
    early_stopping_round: ``int``
        The round of early stop.
    weight_decay: ``float``
        The weight decay argument for optimizer
    device: ``torch.device`` or ``str``
        The device where model will be running on.
    init: ``bool``
        If True(False), the model will (not) be initialized.
    feval: ``str``.
        The evaluation method adopted in this function.
    """

    def __init__(
        self,
        model: _typing.Union[BaseAutoModel, str],
        num_features: int,
        num_classes: int,
        optimizer: _typing.Union[_typing.Type[torch.optim.Optimizer], str, None] = ...,
        lr: float = 1e-4,
        max_epoch: int = 100,
        early_stopping_round: int = 100,
        weight_decay: float = 1e-4,
        device: _typing.Optional[torch.device] = None,
        init: bool = True,
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (MicroF1,),
        loss: str = "nll_loss",
        lr_scheduler_type: _typing.Optional[str] = None,
        **kwargs,
    ) -> None:
        if isinstance(optimizer, type) and issubclass(optimizer, torch.optim.Optimizer):
            self._optimizer_class: _typing.Type[torch.optim.Optimizer] = optimizer
        elif type(optimizer) == str:
            if optimizer.lower() == "adam":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.Adam
            elif optimizer.lower() == "adam" + "w":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.AdamW
            elif optimizer.lower() == "sgd":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.SGD
            else:
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.Adam
        else:
            self._optimizer_class: _typing.Type[
                torch.optim.Optimizer
            ] = torch.optim.Adam
        self._learning_rate: float = lr if lr > 0 else 1e-4
        self._lr_scheduler_type: _typing.Optional[str] = lr_scheduler_type
        self._max_epoch: int = max_epoch if max_epoch > 0 else 1e2
        self._weight_decay: float = weight_decay if weight_decay > 0 else 1e-4
        self._early_stopping = EarlyStopping(
            patience=early_stopping_round if early_stopping_round > 0 else 1e2,
            verbose=False,
        )
        """ Assign an empty initial hyper parameter space """
        self._hyper_parameter_space: _typing.Sequence[
            _typing.Dict[str, _typing.Any]
        ] = []

        self._valid_result: torch.Tensor = torch.zeros(0)
        self._valid_result_prob: torch.Tensor = torch.zeros(0)
        self._valid_score: _typing.Sequence[float] = ()

        """ Set hyper parameters """
        self.__sampled_node_sizes: _typing.Sequence[int] = kwargs.get(
            "sampled_node_sizes"
        )

        self.__training_batch_size: int = kwargs.get("training_batch_size", 1024)
        if not self.__training_batch_size > 0:
            self.__training_batch_size: int = 1024
        self.__predicting_batch_size: int = kwargs.get("predicting_batch_size", 1024)
        if not self.__predicting_batch_size > 0:
            self.__predicting_batch_size: int = 1024

        cpu_count: int = os.cpu_count() if os.cpu_count() is not None else 0
        self.__training_sampler_num_workers: int = kwargs.get(
            "training_sampler_num_workers", cpu_count
        )
        if self.__training_sampler_num_workers > cpu_count:
            self.__training_sampler_num_workers = cpu_count
        self.__predicting_sampler_num_workers: int = kwargs.get(
            "predicting_sampler_num_workers", cpu_count
        )
        if self.__predicting_sampler_num_workers > cpu_count:
            self.__predicting_sampler_num_workers = cpu_count

        super(NodeClassificationLayerDependentImportanceSamplingTrainer, self).__init__(
            model, None, num_features, num_classes, device, feval, loss
        )

        self.__neighbor_sampler_store: _DeterministicNeighborSamplerStore = (
            _DeterministicNeighborSamplerStore()
        )

        self.__is_initialized: bool = False
        if init:
            self.initialize()

    def initialize(self):
        if self.__is_initialized:
            return self
        self.model.initialize()
        self.__is_initialized = True
        return self

    def to(self, device: torch.device):
        self.device = device
        if self.model is not None:
            self.model.to(self.device)

    def get_model(self):
        return self.model

    def __train_only(self, integral_data):
        """
        The function of training on the given dataset and mask.
        :param integral_data: data of a specific graph
        :return: self
        """
        optimizer: torch.optim.Optimizer = self._optimizer_class(
            self.model.model.parameters(),
            lr=self._learning_rate,
            weight_decay=self._weight_decay,
        )

        if type(self._lr_scheduler_type) == str:
            if self._lr_scheduler_type.lower() == "step" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.StepLR = (
                    torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.1)
                )
            elif self._lr_scheduler_type.lower() == "multi" + "step" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.MultiStepLR = (
                    torch.optim.lr_scheduler.MultiStepLR(
                        optimizer, milestones=[30, 80], gamma=0.1
                    )
                )
            elif self._lr_scheduler_type.lower() == "exponential" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.ExponentialLR = (
                    torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.1)
                )
            elif self._lr_scheduler_type.lower() == "ReduceLROnPlateau".lower():
                lr_scheduler: torch.optim.lr_scheduler.ReduceLROnPlateau = (
                    torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, "min")
                )
            else:
                lr_scheduler: torch.optim.lr_scheduler.LambdaLR = (
                    torch.optim.lr_scheduler.LambdaLR(optimizer, lambda _: 1.0)
                )
        else:
            lr_scheduler: torch.optim.lr_scheduler.LambdaLR = (
                torch.optim.lr_scheduler.LambdaLR(optimizer, lambda _: 1.0)
            )

        __layer_dependent_importance_sampler: LayerDependentImportanceSampler = (
            LayerDependentImportanceSampler(
                integral_data.edge_index,
                torch.where(integral_data.train_mask)[0].unique(),
                self.__sampled_node_sizes,
                batch_size=self.__training_batch_size,
                num_workers=self.__training_sampler_num_workers,
            )
        )
        for current_epoch in range(self._max_epoch):
            self.model.model.train()
            optimizer.zero_grad()
            """ epoch start """
            " sample graphs "
            for sampled_data in __layer_dependent_importance_sampler:
                optimizer.zero_grad()
                sampled_data: TargetDependantSampledData = sampled_data
                # 由于现在的Model设计是接受Data的，所以只能组装一个采样的Data作为参数
                sampled_graph: autogl.data.Data = autogl.data.Data(
                    x=integral_data.x[sampled_data.all_sampled_nodes_indexes],
                    y=integral_data.y[sampled_data.all_sampled_nodes_indexes],
                )
                sampled_graph.to(self.device)
                sampled_graph.edge_indexes: _typing.Sequence[torch.LongTensor] = [
                    current_layer.edge_index_for_sampled_graph.to(self.device)
                    for current_layer in sampled_data.sampled_edges_for_layers
                ]
                sampled_graph.edge_weights: _typing.Sequence[torch.Tensor] = [
                    current_layer.edge_weight.to(self.device)
                    for current_layer in sampled_data.sampled_edges_for_layers
                ]
                if isinstance(self.model.model, ClassificationSupportedSequentialModel):
                    prediction: torch.Tensor = self.model.model.cls_forward(
                        sampled_graph
                    )
                else:
                    prediction: torch.Tensor = self.model.model(sampled_graph)
                if not hasattr(torch.nn.functional, self.loss):
                    raise TypeError(f"PyTorch does not support loss type {self.loss}")
                loss_function = getattr(torch.nn.functional, self.loss)
                loss_value: torch.Tensor = loss_function(
                    prediction[
                        sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                    ],
                    sampled_graph.y[
                        sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                    ],
                )
                loss_value.backward()
                optimizer.step()

            if self._lr_scheduler_type:
                lr_scheduler.step()

            if (
                hasattr(integral_data, "val_mask")
                and getattr(integral_data, "val_mask") is not None
                and type(getattr(integral_data, "val_mask")) == torch.Tensor
            ):
                validation_results: _typing.Sequence[float] = self.evaluate(
                    (integral_data,), "val", [self.feval[0]]
                )
                if self.feval[0].is_higher_better():
                    validation_loss: float = -validation_results[0]
                else:
                    validation_loss: float = validation_results[0]
                self._early_stopping(validation_loss, self.model.model)
                if self._early_stopping.early_stop:
                    LOGGER.debug("Early stopping at %d", current_epoch)
                    break
        if (
            hasattr(integral_data, "val_mask")
            and getattr(integral_data, "val_mask") is not None
            and type(getattr(integral_data, "val_mask")) == torch.Tensor
        ):
            self._early_stopping.load_checkpoint(self.model.model)

    def __predict_only(
        self,
        integral_data,
        mask_or_target_nodes_indexes: _typing.Union[torch.BoolTensor, torch.LongTensor],
    ) -> torch.Tensor:
        """
        The function of predicting on the given data.
        :param integral_data: data of a specific graph
        :param mask_or_target_nodes_indexes: ...
        :return: the result of prediction on the given dataset
        """
        self.model.model.eval()
        integral_data = integral_data.to(torch.device("cpu"))
        mask_or_target_nodes_indexes = mask_or_target_nodes_indexes.to(
            torch.device("cpu")
        )
        if isinstance(self.model.model, ClassificationSupportedSequentialModel):
            sequential_gnn_model: ClassificationSupportedSequentialModel = (
                self.model.model
            )
            __num_layers: int = len(self.__sampled_node_sizes)

            x: torch.Tensor = getattr(integral_data, "x")
            for _current_layer_index in range(__num_layers - 1):
                __next_x: _typing.Optional[torch.Tensor] = None

                _optional_neighbor_sampler: _typing.Optional[
                    NeighborSampler
                ] = self.__neighbor_sampler_store[torch.arange(x.size(0))]
                if (
                    _optional_neighbor_sampler is not None
                    and type(_optional_neighbor_sampler) == NeighborSampler
                ):
                    current_neighbor_sampler: NeighborSampler = (
                        _optional_neighbor_sampler
                    )
                else:
                    current_neighbor_sampler: NeighborSampler = NeighborSampler(
                        integral_data.edge_index,
                        torch.arange(x.size(0)).unique(),
                        [-1],
                        batch_size=self.__predicting_batch_size,
                        num_workers=self.__predicting_sampler_num_workers,
                        shuffle=False,
                    )
                    self.__neighbor_sampler_store[
                        torch.arange(x.size(0))
                    ] = current_neighbor_sampler

                for _target_dependant_sampled_data in current_neighbor_sampler:
                    _target_dependant_sampled_data: TargetDependantSampledData = (
                        _target_dependant_sampled_data
                    )
                    _sampled_graph: autogl.data.Data = autogl.data.Data(
                        x=x[_target_dependant_sampled_data.all_sampled_nodes_indexes],
                        edge_index=(
                            _target_dependant_sampled_data.sampled_edges_for_layers[
                                0
                            ].edge_index_for_sampled_graph
                        ),
                    )
                    _sampled_graph.edge_weight: torch.Tensor = (
                        _target_dependant_sampled_data.sampled_edges_for_layers[
                            0
                        ].edge_weight
                    )
                    _sampled_graph: autogl.data.Data = _sampled_graph.to(self.device)

                    with torch.no_grad():
                        __sampled_graph_inferences: torch.Tensor = (
                            sequential_gnn_model.sequential_encoding_layers[
                                _current_layer_index
                            ](_sampled_graph)
                        )
                        _sampled_target_nodes_inferences: torch.Tensor = __sampled_graph_inferences[
                            _target_dependant_sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                        ].cpu()
                    if __next_x is None:
                        __next_x: torch.Tensor = torch.zeros(
                            x.size(0), __sampled_graph_inferences.size(1)
                        )
                    __next_x[
                        _target_dependant_sampled_data.target_nodes_indexes.indexes_in_integral_graph
                    ] = _sampled_target_nodes_inferences
                x: torch.Tensor = __next_x
            " The following procedures are for the top layer "
            if mask_or_target_nodes_indexes.dtype == torch.bool:
                target_nodes_indexes: _typing.Any = torch.where(
                    mask_or_target_nodes_indexes
                )[0]
            else:
                target_nodes_indexes: _typing.Any = mask_or_target_nodes_indexes.long()

            _optional_neighbor_sampler: _typing.Optional[
                NeighborSampler
            ] = self.__neighbor_sampler_store[target_nodes_indexes]
            if (
                _optional_neighbor_sampler is not None
                and type(_optional_neighbor_sampler) == NeighborSampler
            ):
                current_neighbor_sampler: NeighborSampler = _optional_neighbor_sampler
            else:
                current_neighbor_sampler: NeighborSampler = NeighborSampler(
                    integral_data.edge_index,
                    target_nodes_indexes,
                    [-1],
                    batch_size=self.__predicting_batch_size,
                    num_workers=self.__predicting_sampler_num_workers,
                    shuffle=False,
                )
                self.__neighbor_sampler_store[
                    target_nodes_indexes
                ] = current_neighbor_sampler

            prediction_batch_cumulative_builder = (
                EvaluatorUtility.PredictionBatchCumulativeBuilder()
            )
            for _target_dependant_sampled_data in current_neighbor_sampler:
                _sampled_graph: autogl.data.Data = autogl.data.Data(
                    x[_target_dependant_sampled_data.all_sampled_nodes_indexes],
                    _target_dependant_sampled_data.sampled_edges_for_layers[
                        0
                    ].edge_index_for_sampled_graph,
                )
                _sampled_graph.edge_weight: torch.Tensor = (
                    _target_dependant_sampled_data.sampled_edges_for_layers[
                        0
                    ].edge_weight
                )
                _sampled_graph: autogl.data.Data = _sampled_graph.to(self.device)
                with torch.no_grad():
                    prediction_batch_cumulative_builder.add_batch(
                        _target_dependant_sampled_data.target_nodes_indexes.indexes_in_integral_graph.cpu().numpy(),
                        sequential_gnn_model.cls_decode(
                            sequential_gnn_model.sequential_encoding_layers[-1](
                                _sampled_graph
                            )
                        )[
                            _target_dependant_sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                        ]
                        .cpu()
                        .numpy(),
                    )
            return torch.from_numpy(prediction_batch_cumulative_builder.compose()[1])
        else:
            if mask_or_target_nodes_indexes.dtype == torch.bool:
                target_nodes_indexes: _typing.Any = torch.where(
                    mask_or_target_nodes_indexes
                )[0]
            else:
                target_nodes_indexes: _typing.Any = mask_or_target_nodes_indexes.long()
            neighbor_sampler: NeighborSampler = NeighborSampler(
                torch_geometric.utils.add_remaining_self_loops(
                    integral_data.edge_index
                )[0],
                target_nodes_indexes,
                [-1 for _ in self.__sampled_node_sizes],
                batch_size=self.__predicting_batch_size,
                num_workers=self.__predicting_sampler_num_workers,
                shuffle=False,
            )
            prediction_batch_cumulative_builder = (
                EvaluatorUtility.PredictionBatchCumulativeBuilder()
            )
            self.model.model.eval()
            for sampled_data in neighbor_sampler:
                sampled_data: TargetDependantSampledData = sampled_data
                sampled_graph: autogl.data.Data = autogl.data.Data(
                    integral_data.x[sampled_data.all_sampled_nodes_indexes],
                    integral_data.y[sampled_data.all_sampled_nodes_indexes],
                )
                sampled_graph.to(self.device)
                sampled_graph.edge_indexes: _typing.Sequence[torch.LongTensor] = [
                    current_layer.edge_index_for_sampled_graph.to(self.device)
                    for current_layer in sampled_data.sampled_edges_for_layers
                ]
                sampled_graph.edge_weights: _typing.Sequence[torch.FloatTensor] = [
                    current_layer.edge_weight.to(self.device)
                    for current_layer in sampled_data.sampled_edges_for_layers
                ]
                with torch.no_grad():
                    prediction_batch_cumulative_builder.add_batch(
                        sampled_data.target_nodes_indexes.indexes_in_integral_graph.cpu().numpy(),
                        self.model.model(sampled_graph)[
                            sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                        ]
                        .cpu()
                        .numpy(),
                    )
            return torch.from_numpy(prediction_batch_cumulative_builder.compose()[1])

    def predict_proba(
        self, dataset, mask: _typing.Optional[str] = None, in_log_format: bool = False
    ):
        """
        The function of predicting the probability on the given dataset.

        Parameters
        ----------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.
        in_log_format: ``bool``.
            If True(False), the probability will (not) be log format.

        Returns
        -------
        The prediction result.
        """
        data = dataset[0].to(torch.device("cpu"))
        if mask is not None and type(mask) == str:
            if mask.lower() == "train":
                _mask: torch.BoolTensor = data.train_mask
            elif mask.lower() == "test":
                _mask: torch.BoolTensor = data.test_mask
            elif mask.lower() == "val":
                _mask: torch.BoolTensor = data.val_mask
            else:
                _mask: torch.BoolTensor = data.test_mask
        else:
            _mask: torch.BoolTensor = data.test_mask
        result = self.__predict_only(data, _mask)
        return result if in_log_format else torch.exp(result)

    def predict(self, dataset, mask: _typing.Optional[str] = None) -> torch.Tensor:
        """
        The function of predicting on the given dataset.

        Parameters
        ----------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.

        Returns
        -------
        The prediction result of ``predict_proba``.
        """
        return self.predict_proba(dataset, mask, in_log_format=True).max(1)[1]

    def evaluate(
        self,
        dataset,
        mask: _typing.Optional[str] = None,
        feval: _typing.Union[
            None, _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = None,
    ) -> _typing.Sequence[float]:
        """
        The function of training on the given dataset and keeping valid result.

        Parameters
        ------------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.
        feval: ``str``.
            The evaluation method adopted in this function.

        Returns
        -------
        result: The evaluation result on the given dataset.
        """
        data = dataset[0]
        data = data.to(self.device)
        if feval is None:
            _feval: _typing.Sequence[_typing.Type[Evaluation]] = self.feval
        else:
            _feval: _typing.Sequence[_typing.Type[Evaluation]] = get_feval(list(feval))
        if mask is not None and type(mask) == str:
            if mask.lower() == "train":
                _mask: torch.Tensor = data.train_mask
            elif mask.lower() == "test":
                _mask: torch.Tensor = data.test_mask
            elif mask.lower() == "val":
                _mask: torch.Tensor = data.val_mask
            else:
                _mask: torch.Tensor = data.test_mask
        else:
            _mask: torch.Tensor = data.test_mask
        prediction_probability: torch.Tensor = self.predict_proba(dataset, mask)
        y_ground_truth: torch.Tensor = data.y[_mask]

        return [
            f.evaluate(
                prediction_probability.cpu().numpy(),
                y_ground_truth.cpu().numpy(),
            )
            for f in _feval
        ]

    def train(self, dataset, keep_valid_result: bool = True):
        """
        The function of training on the given dataset and keeping valid result.

        Parameters
        ------------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        keep_valid_result: ``bool``
            If True(False), save the validation result after training.

        Returns
        --------
        None
        """
        import gc

        gc.collect()
        data = dataset[0]
        self.__train_only(data)
        if keep_valid_result:
            data = data.to(torch.device("cpu"))
            prediction: torch.Tensor = self.__predict_only(data, data.val_mask)
            self._valid_result: torch.Tensor = prediction.max(1)[1]
            self._valid_result_prob: torch.Tensor = prediction
            self._valid_score: _typing.Sequence[float] = self.evaluate(dataset, "val")

    def get_valid_predict(self) -> torch.Tensor:
        return self._valid_result

    def get_valid_predict_proba(self) -> torch.Tensor:
        return self._valid_result_prob

    def get_valid_score(
        self, return_major: bool = True
    ) -> _typing.Union[
        _typing.Tuple[float, bool],
        _typing.Tuple[_typing.Sequence[float], _typing.Sequence[bool]],
    ]:
        """
        The function of getting the valid score.

        Parameters
        ----------
        return_major: ``bool``.
            If True, the return only consists of the major result.
            If False, the return consists of the all results.

        Returns
        -------
        result: The valid score.
        """
        if return_major:
            return self._valid_score[0], self.feval[0].is_higher_better()
        else:
            return self._valid_score, [f.is_higher_better() for f in self.feval]

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(
            {
                "trainer_name": self.__class__.__name__,
                "optimizer": self._optimizer_class,
                "learning_rate": self._learning_rate,
                "max_epoch": self._max_epoch,
                "early_stopping_round": self._early_stopping.patience,
                "sampling_sizes": self.__sampled_node_sizes,
                "model": repr(self.model),
            }
        )

    def duplicate_from_hyper_parameter(
        self,
        hp: _typing.Dict[str, _typing.Any],
        model: _typing.Optional[BaseAutoModel] = None,
    ) -> "NodeClassificationLayerDependentImportanceSamplingTrainer":
        """
        The function of duplicating a new instance from the given hyper-parameter.

        Parameters
        ------------
        hp: ``dict``.
            The hyper-parameter settings for the new instance.
        model: ``BaseModel``
            The name or class of model adopted

        Returns
        --------
        instance: ``NodeClassificationLayerDependentImportanceSamplingTrainer``
            A new instance of trainer.
        """
        if model is None or not isinstance(model, BaseAutoModel):
            model: BaseAutoModel = self.model
        model = model.from_hyper_parameter(hp["encoder"])
        trainer_hp = hp["trainer"]
        return NodeClassificationLayerDependentImportanceSamplingTrainer(
            model,
            self.num_features,
            self.num_classes,
            self._optimizer_class,
            device=self.device,
            init=True,
            feval=self.feval,
            loss=self.loss,
            lr_scheduler_type=self._lr_scheduler_type,
            **trainer_hp,
        )


@register_trainer("NodeClassificationNeighborSamplingTrainer")
class NodeClassificationNeighborSamplingTrainer(BaseNodeClassificationTrainer):
    """
    The node classification trainer utilizing Layer dependent importance sampling technique.

    Parameters
    ------------
    model: ``BaseModel`` or ``str``
        The name or class of model adopted
    num_features: ``int``
        number of features for each node provided by dataset
    num_classes: ``int``
        number of classes to classify
    optimizer: ``Optimizer`` of ``str``
        The (name of) optimizer used to train and predict.
    lr: ``float``
        The learning rate of link prediction task.
    max_epoch: ``int``
        The max number of epochs in training.
    early_stopping_round: ``int``
        The round of early stop.
    weight_decay: ``float``
        The weight decay argument for optimizer
    device: ``torch.device`` or ``str``
        The device where model will be running on.
    init: ``bool``
        If True(False), the model will (not) be initialized.
    feval: ``str``.
        The evaluation method adopted in this function.
    """

    def __init__(
        self,
        model: _typing.Union[BaseAutoModel, str],
        num_features: int,
        num_classes: int,
        optimizer: _typing.Union[_typing.Type[torch.optim.Optimizer], str, None] = ...,
        lr: float = 1e-4,
        max_epoch: int = 100,
        early_stopping_round: int = 100,
        weight_decay: float = 1e-4,
        device: _typing.Optional[torch.device] = None,
        init: bool = True,
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (Logloss,),
        loss: str = "nll_loss",
        lr_scheduler_type: _typing.Optional[str] = None,
        **kwargs,
    ):
        if isinstance(optimizer, type) and issubclass(optimizer, torch.optim.Optimizer):
            self._optimizer_class: _typing.Type[torch.optim.Optimizer] = optimizer
        elif type(optimizer) == str:
            if optimizer.lower() == "adam":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.Adam
            elif optimizer.lower() == "adam" + "w":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.AdamW
            elif optimizer.lower() == "sgd":
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.SGD
            else:
                self._optimizer_class: _typing.Type[
                    torch.optim.Optimizer
                ] = torch.optim.Adam
        else:
            self._optimizer_class: _typing.Type[
                torch.optim.Optimizer
            ] = torch.optim.Adam
        self._learning_rate: float = lr if lr > 0 else 1e-4
        self._lr_scheduler_type: _typing.Optional[str] = lr_scheduler_type
        self._max_epoch: int = max_epoch if max_epoch > 0 else 1e2
        self._weight_decay: float = weight_decay if weight_decay > 0 else 1e-4
        self._early_stopping = EarlyStopping(
            patience=early_stopping_round if early_stopping_round > 0 else 1e2,
            verbose=False,
        )
        """ Assign an empty initial hyper parameter space """
        self._hyper_parameter_space: _typing.Sequence[
            _typing.Dict[str, _typing.Any]
        ] = []

        self._valid_result: torch.Tensor = torch.zeros(0)
        self._valid_result_prob: torch.Tensor = torch.zeros(0)
        self._valid_score: _typing.Sequence[float] = ()

        """ Set hyper-parameter """
        self.__sampling_sizes: _typing.Sequence[int] = kwargs.get("sampling_sizes")

        self.__training_batch_size: int = kwargs.get("training_batch_size", 1024)
        if not self.__training_batch_size > 0:
            self.__training_batch_size: int = 1024
        self.__predicting_batch_size: int = kwargs.get("predicting_batch_size", 1024)
        if not self.__predicting_batch_size > 0:
            self.__predicting_batch_size: int = 1024

        cpu_count: int = os.cpu_count() if os.cpu_count() is not None else 0
        self.__training_sampler_num_workers: int = kwargs.get(
            "training_sampler_num_workers", cpu_count
        )
        if self.__training_sampler_num_workers > cpu_count:
            self.__training_sampler_num_workers = cpu_count
        self.__predicting_sampler_num_workers: int = kwargs.get(
            "predicting_sampler_num_workers", cpu_count
        )
        if self.__predicting_sampler_num_workers > cpu_count:
            self.__predicting_sampler_num_workers = cpu_count

        super(NodeClassificationNeighborSamplingTrainer, self).__init__(
            model, None, num_features, num_classes, device, feval, loss
        )

        self.__neighbor_sampler_store: _DeterministicNeighborSamplerStore = (
            _DeterministicNeighborSamplerStore()
        )

        self.__is_initialized: bool = False
        if init:
            self.initialize()

    def initialize(self):
        if self.__is_initialized:
            return self
        self.model.initialize()
        self.__is_initialized = True
        return self

    def to(self, device: torch.device):
        self.device = device
        if self.model is not None:
            self.model.to(self.device)

    def get_model(self):
        return self.model

    def __train_only(self, integral_data):
        """
        The function of training on the given dataset and mask.
        :param integral_data: data of the integral graph
        :return: self
        """
        optimizer: torch.optim.Optimizer = self._optimizer_class(
            self.model.model.parameters(),
            lr=self._learning_rate,
            weight_decay=self._weight_decay,
        )

        if type(self._lr_scheduler_type) == str:
            if self._lr_scheduler_type.lower() == "step" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.StepLR = (
                    torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.1)
                )
            elif self._lr_scheduler_type.lower() == "multi" + "step" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.MultiStepLR = (
                    torch.optim.lr_scheduler.MultiStepLR(
                        optimizer, milestones=[30, 80], gamma=0.1
                    )
                )
            elif self._lr_scheduler_type.lower() == "exponential" + "lr":
                lr_scheduler: torch.optim.lr_scheduler.ExponentialLR = (
                    torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.1)
                )
            elif self._lr_scheduler_type.lower() == "ReduceLROnPlateau".lower():
                lr_scheduler: torch.optim.lr_scheduler.ReduceLROnPlateau = (
                    torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, "min")
                )
            else:
                lr_scheduler: torch.optim.lr_scheduler.LambdaLR = (
                    torch.optim.lr_scheduler.LambdaLR(optimizer, lambda _: 1.0)
                )
        else:
            lr_scheduler: torch.optim.lr_scheduler.LambdaLR = (
                torch.optim.lr_scheduler.LambdaLR(optimizer, lambda _: 1.0)
            )

        neighbor_sampler: NeighborSampler = NeighborSampler(
            integral_data.edge_index,
            torch.where(integral_data.train_mask)[0].unique(),
            self.__sampling_sizes,
            batch_size=self.__training_batch_size,
            num_workers=self.__training_sampler_num_workers,
        )
        for current_epoch in range(self._max_epoch):
            self.model.model.train()
            optimizer.zero_grad()
            """ epoch start """
            " sample graphs "
            for sampled_data in neighbor_sampler:
                optimizer.zero_grad()
                sampled_data: TargetDependantSampledData = sampled_data
                sampled_graph: autogl.data.Data = autogl.data.Data(
                    x=integral_data.x[sampled_data.all_sampled_nodes_indexes],
                    y=integral_data.y[sampled_data.all_sampled_nodes_indexes],
                )
                sampled_graph.to(self.device)
                sampled_graph.edge_indexes: _typing.Sequence[torch.LongTensor] = [
                    current_layer.edge_index_for_sampled_graph.to(self.device)
                    for current_layer in sampled_data.sampled_edges_for_layers
                ]
                if isinstance(self.model.model, ClassificationSupportedSequentialModel):
                    prediction: torch.Tensor = self.model.model.cls_forward(
                        sampled_graph
                    )
                else:
                    prediction: torch.Tensor = self.model.model(sampled_graph)
                if not hasattr(torch.nn.functional, self.loss):
                    raise TypeError(f"PyTorch does not support loss type {self.loss}")
                loss_function = getattr(torch.nn.functional, self.loss)
                loss_value: torch.Tensor = loss_function(
                    prediction[
                        sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                    ],
                    sampled_graph.y[
                        sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                    ],
                )
                loss_value.backward()
                optimizer.step()

            if self._lr_scheduler_type:
                lr_scheduler.step()

            if (
                hasattr(integral_data, "val_mask")
                and getattr(integral_data, "val_mask") is not None
                and type(getattr(integral_data, "val_mask")) == torch.Tensor
            ):
                validation_results: _typing.Sequence[float] = self.evaluate(
                    (integral_data,), "val", [self.feval[0]]
                )
                if self.feval[0].is_higher_better():
                    validation_loss: float = -validation_results[0]
                else:
                    validation_loss: float = validation_results[0]
                self._early_stopping(validation_loss, self.model.model)
                if self._early_stopping.early_stop:
                    LOGGER.debug("Early stopping at %d", current_epoch)
                    break
        if (
            hasattr(integral_data, "val_mask")
            and getattr(integral_data, "val_mask") is not None
            and type(getattr(integral_data, "val_mask")) == torch.Tensor
        ):
            self._early_stopping.load_checkpoint(self.model.model)

    def __predict_only(
        self,
        integral_data,
        mask_or_target_nodes_indexes: _typing.Union[torch.BoolTensor, torch.LongTensor],
    ) -> torch.Tensor:
        """
        The function of predicting on the given data.
        :param integral_data: data of a specific graph
        :param mask_or_target_nodes_indexes: ...
        :return: the result of prediction on the given dataset
        """
        self.model.model.eval()
        integral_data = integral_data.to(torch.device("cpu"))
        mask_or_target_nodes_indexes = mask_or_target_nodes_indexes.to(
            torch.device("cpu")
        )
        if isinstance(self.model.model, ClassificationSupportedSequentialModel):
            sequential_gnn_model: ClassificationSupportedSequentialModel = (
                self.model.model
            )
            __num_layers: int = len(self.__sampling_sizes)

            x: torch.Tensor = getattr(integral_data, "x")
            for _current_layer_index in range(__num_layers - 1):
                __next_x: _typing.Optional[torch.Tensor] = None

                _optional_neighbor_sampler: _typing.Optional[
                    NeighborSampler
                ] = self.__neighbor_sampler_store[torch.arange(x.size(0)).unique()]
                if (
                    _optional_neighbor_sampler is not None
                    and type(_optional_neighbor_sampler) == NeighborSampler
                ):
                    current_neighbor_sampler: NeighborSampler = (
                        _optional_neighbor_sampler
                    )
                else:
                    current_neighbor_sampler: NeighborSampler = NeighborSampler(
                        integral_data.edge_index,
                        torch.arange(x.size(0)).unique(),
                        [-1],
                        batch_size=self.__predicting_batch_size,
                        num_workers=self.__predicting_sampler_num_workers,
                        shuffle=False,
                    )
                    __temp: _typing.Any = torch.arange(x.size(0))
                    self.__neighbor_sampler_store[__temp] = current_neighbor_sampler

                for _target_dependant_sampled_data in current_neighbor_sampler:
                    _target_dependant_sampled_data: TargetDependantSampledData = (
                        _target_dependant_sampled_data
                    )
                    _sampled_graph: autogl.data.Data = autogl.data.Data(
                        x=x[_target_dependant_sampled_data.all_sampled_nodes_indexes],
                        edge_index=(
                            _target_dependant_sampled_data.sampled_edges_for_layers[
                                0
                            ].edge_index_for_sampled_graph
                        ),
                    )
                    _sampled_graph: autogl.data.Data = _sampled_graph.to(self.device)

                    with torch.no_grad():
                        __sampled_graph_inferences: torch.Tensor = (
                            sequential_gnn_model.sequential_encoding_layers[
                                _current_layer_index
                            ](_sampled_graph)
                        )
                        _sampled_target_nodes_inferences: torch.Tensor = __sampled_graph_inferences[
                            _target_dependant_sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                        ].cpu()
                    if __next_x is None:
                        __next_x: torch.Tensor = torch.zeros(
                            x.size(0), __sampled_graph_inferences.size(1)
                        )
                    __next_x[
                        _target_dependant_sampled_data.target_nodes_indexes.indexes_in_integral_graph
                    ] = _sampled_target_nodes_inferences
                x: torch.Tensor = __next_x
            # The following procedures are for the top layer
            if mask_or_target_nodes_indexes.dtype == torch.bool:
                target_nodes_indexes: _typing.Any = torch.where(
                    mask_or_target_nodes_indexes
                )[0]
            else:
                target_nodes_indexes: _typing.Any = mask_or_target_nodes_indexes.long()

            _optional_neighbor_sampler: _typing.Optional[
                NeighborSampler
            ] = self.__neighbor_sampler_store[target_nodes_indexes]
            if (
                _optional_neighbor_sampler is not None
                and type(_optional_neighbor_sampler) == NeighborSampler
            ):
                current_neighbor_sampler: NeighborSampler = _optional_neighbor_sampler
            else:
                current_neighbor_sampler: NeighborSampler = NeighborSampler(
                    integral_data.edge_index,
                    target_nodes_indexes,
                    [-1],
                    batch_size=self.__predicting_batch_size,
                    num_workers=self.__predicting_sampler_num_workers,
                    shuffle=False,
                )
                self.__neighbor_sampler_store[
                    target_nodes_indexes
                ] = current_neighbor_sampler

            prediction_batch_cumulative_builder = (
                EvaluatorUtility.PredictionBatchCumulativeBuilder()
            )
            for _target_dependant_sampled_data in current_neighbor_sampler:
                _sampled_graph: autogl.data.Data = autogl.data.Data(
                    x[_target_dependant_sampled_data.all_sampled_nodes_indexes],
                    _target_dependant_sampled_data.sampled_edges_for_layers[
                        0
                    ].edge_index_for_sampled_graph,
                )
                _sampled_graph: autogl.data.Data = _sampled_graph.to(self.device)
                with torch.no_grad():
                    prediction_batch_cumulative_builder.add_batch(
                        _target_dependant_sampled_data.target_nodes_indexes.indexes_in_integral_graph.cpu().numpy(),
                        sequential_gnn_model.cls_decode(
                            sequential_gnn_model.sequential_encoding_layers[-1](
                                _sampled_graph
                            )
                        )[
                            _target_dependant_sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                        ]
                        .cpu()
                        .numpy(),
                    )
            return torch.from_numpy(prediction_batch_cumulative_builder.compose()[1])
        else:
            if mask_or_target_nodes_indexes.dtype == torch.bool:
                target_nodes_indexes: _typing.Any = torch.where(
                    mask_or_target_nodes_indexes
                )[0]
            else:
                target_nodes_indexes: _typing.Any = mask_or_target_nodes_indexes.long()

            neighbor_sampler: NeighborSampler = NeighborSampler(
                integral_data.edge_index,
                target_nodes_indexes,
                [-1 for _ in self.__sampling_sizes],
                batch_size=self.__predicting_batch_size,
                num_workers=self.__predicting_sampler_num_workers,
                shuffle=False,
            )

            prediction_batch_cumulative_builder = (
                EvaluatorUtility.PredictionBatchCumulativeBuilder()
            )
            self.model.model.eval()
            for _target_dependant_sampled_data in neighbor_sampler:
                _sampled_graph: autogl.data.Data = autogl.data.Data(
                    x=integral_data.x[
                        _target_dependant_sampled_data.all_sampled_nodes_indexes
                    ]
                )
                _sampled_graph = _sampled_graph.to(self.device)
                _sampled_graph.edge_indexes: _typing.Sequence[torch.LongTensor] = [
                    current_layer.edge_index_for_sampled_graph.to(self.device)
                    for current_layer in _target_dependant_sampled_data.sampled_edges_for_layers
                ]
                with torch.no_grad():
                    prediction_batch_cumulative_builder.add_batch(
                        _target_dependant_sampled_data.target_nodes_indexes.indexes_in_integral_graph.cpu().numpy(),
                        self.model.model(_sampled_graph)[
                            _target_dependant_sampled_data.target_nodes_indexes.indexes_in_sampled_graph
                        ]
                        .cpu()
                        .numpy(),
                    )
            return torch.from_numpy(prediction_batch_cumulative_builder.compose()[1])

    def predict_proba(
        self, dataset, mask: _typing.Optional[str] = None, in_log_format: bool = False
    ):
        """
        The function of predicting the probability on the given dataset.

        Parameters
        ----------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.
        in_log_format: ``bool``.
            If True(False), the probability will (not) be log format.

        Returns
        -------
        The prediction result.
        """
        data = dataset[0].to(torch.device("cpu"))
        if mask is not None and type(mask) == str:
            if mask.lower() == "train":
                _mask: torch.BoolTensor = data.train_mask
            elif mask.lower() == "test":
                _mask: torch.BoolTensor = data.test_mask
            elif mask.lower() == "val":
                _mask: torch.BoolTensor = data.val_mask
            else:
                _mask: torch.BoolTensor = data.test_mask
        else:
            _mask: torch.BoolTensor = data.test_mask
        result = self.__predict_only(data, _mask)
        return result if in_log_format else torch.exp(result)

    def predict(self, dataset, mask: _typing.Optional[str] = None) -> torch.Tensor:
        """
        The function of predicting on the given dataset.

        Parameters
        ----------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.

        Returns
        -------
        The prediction result of ``predict_proba``.
        """
        return self.predict_proba(dataset, mask, in_log_format=True).max(1)[1]

    def evaluate(
        self,
        dataset,
        mask: _typing.Optional[str] = None,
        feval: _typing.Union[
            None, _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = None,
    ) -> _typing.Sequence[float]:
        """
        The function of training on the given dataset and keeping valid result.

        Parameters
        ------------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        mask: .
            The `str` of ``train``, ``val``, or ``test``,
            representing the identifier for specific dataset mask.
        feval: ``str``.
            The evaluation method adopted in this function.

        Returns
        -------
        result: The evaluation result on the given dataset.
        """
        data = dataset[0]
        data = data.to(self.device)
        if feval is None:
            _feval: _typing.Sequence[_typing.Type[Evaluation]] = self.feval
        else:
            _feval: _typing.Sequence[_typing.Type[Evaluation]] = get_feval(list(feval))
        if mask is not None and type(mask) == str:
            if mask.lower() == "train":
                _mask: torch.Tensor = data.train_mask
            elif mask.lower() == "test":
                _mask: torch.Tensor = data.test_mask
            elif mask.lower() == "val":
                _mask: torch.Tensor = data.val_mask
            else:
                _mask: torch.Tensor = data.test_mask
        else:
            _mask: torch.Tensor = data.test_mask
        prediction_probability: torch.Tensor = self.predict_proba(dataset, mask)
        y_ground_truth: torch.Tensor = data.y[_mask]

        eval_results = []
        for f in _feval:
            eval_results.append(
                f.evaluate(
                    prediction_probability.cpu().numpy(),
                    y_ground_truth.cpu().numpy(),
                )
            )
        return eval_results

    def train(self, dataset, keep_valid_result: bool = True):
        """
        The function of training on the given dataset and keeping valid result.

        Parameters
        ------------
        dataset:
            The dataset containing conventional data of integral graph
            adopted to train for node classification.
        keep_valid_result: ``bool``
            If True(False), save the validation result after training.

        Returns
        --------
        None
        """
        import gc

        gc.collect()
        data = dataset[0]
        self.__train_only(data)
        if keep_valid_result:
            data = data.to(torch.device("cpu"))
            prediction: torch.Tensor = self.__predict_only(data, data.val_mask)
            self._valid_result: torch.Tensor = prediction.max(1)[1]
            self._valid_result_prob: torch.Tensor = prediction
            self._valid_score: _typing.Sequence[float] = self.evaluate(dataset, "val")

    def get_valid_predict(self) -> torch.Tensor:
        return self._valid_result

    def get_valid_predict_proba(self) -> torch.Tensor:
        return self._valid_result_prob

    def get_valid_score(
        self, return_major: bool = True
    ) -> _typing.Union[
        _typing.Tuple[float, bool],
        _typing.Tuple[_typing.Sequence[float], _typing.Sequence[bool]],
    ]:
        """
        The function of getting the valid score.

        Parameters
        ----------
        return_major: ``bool``.
            If True, the return only consists of the major result.
            If False, the return consists of the all results.

        Returns
        -------
        result: The valid score.
        """
        if return_major:
            return self._valid_score[0], self.feval[0].is_higher_better()
        else:
            return self._valid_score, [f.is_higher_better() for f in self.feval]

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(
            {
                "trainer_name": self.__class__.__name__,
                "optimizer": self._optimizer_class,
                "learning_rate": self._learning_rate,
                "max_epoch": self._max_epoch,
                "early_stopping_round": self._early_stopping.patience,
                "sampling_sizes": self.__sampling_sizes,
                "model": repr(self.model),
            }
        )

    def duplicate_from_hyper_parameter(
        self,
        hp: _typing.Dict[str, _typing.Any],
        model: _typing.Optional[BaseAutoModel] = None,
    ) -> "NodeClassificationNeighborSamplingTrainer":
        """
        The function of duplicating a new instance from the given hyper-parameter.

        Parameters
        ------------
        hp: ``dict``.
            The hyper-parameter settings for the new instance.
        model: ``BaseModel``
            The name or class of model adopted

        Returns
        --------
        instance: ``NodeClassificationLayerDependentImportanceSamplingTrainer``
            A new instance of trainer.
        """
        if model is None or not isinstance(model, BaseAutoModel):
            model: BaseAutoModel = self.model
        model = model.from_hyper_parameter(hp["encoder"])
        return NodeClassificationNeighborSamplingTrainer(
            model,
            self.num_features,
            self.num_classes,
            self._optimizer_class,
            device=self.device,
            init=True,
            feval=self.feval,
            loss=self.loss,
            lr_scheduler_type=self._lr_scheduler_type,
            **hp["trainer"],
        )



import torch.nn.functional
import typing as _typing


def activation_func(
        tensor: torch.Tensor, function_name: _typing.Optional[str]
) -> torch.Tensor:
    if not isinstance(function_name, str):
        return tensor
    elif function_name == 'linear':
        return tensor
    elif function_name == 'tanh':
        return torch.tanh(tensor)
    elif hasattr(torch.nn.functional, function_name):
        return getattr(torch.nn.functional, function_name)(tensor)
    else:
        raise TypeError(f"PyTorch does not support activation function {function_name}")



from ...suggestion.algorithm.base_chocolate_algorithm import BaseChocolateAlgorithm


class MocmaesAlgorithm(BaseChocolateAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    Get the new suggested trials with MOCMAES algorithm.
    """

    def __init__(self):
        super(MocmaesAlgorithm, self).__init__("MOCMAES")



"""
Node classification Full Trainer Implementation
"""

from . import register_trainer

from .base import BaseNodeClassificationTrainer, EarlyStopping
import torch
from torch.optim.lr_scheduler import (
    StepLR,
    MultiStepLR,
    ExponentialLR,
    ReduceLROnPlateau,
)
import torch.nn.functional as F
from ..model import BaseEncoderMaintainer, BaseDecoderMaintainer, BaseAutoModel
from .evaluation import Evaluation, get_feval, Logloss
from typing import Callable, Iterable, Optional, Tuple, Type, Union
from copy import deepcopy

from ...utils import get_logger

from ...backend import DependentBackend

LOGGER = get_logger("node classification trainer")


@register_trainer("NodeClassificationFull")
class NodeClassificationFullTrainer(BaseNodeClassificationTrainer):
    """
    The node classification trainer.

    Used to automatically train the node classification problem.

    Parameters
    ----------
    model:
        Models can be ``str``, ``autogl.module.model.BaseAutoModel``, 
        ``autogl.module.model.encoders.BaseEncoderMaintainer`` or a tuple of (encoder, decoder) 
        if need to specify both encoder and decoder. Encoder can be ``str`` or
        ``autogl.module.model.encoders.BaseEncoderMaintainer``, and decoder can be ``str``
        or ``autogl.module.model.decoders.BaseDecoderMaintainer``.
        If only encoder is specified, decoder will be default to "logsoftmax"

    num_features: int (Optional)
        The number of features in dataset. default None
    
    num_classes: int (Optional)
        The number of classes. default None

    optimizer: ``Optimizer`` of ``str``
        The (name of) optimizer used to train and predict. default torch.optim.Adam

    lr: ``float``
        The learning rate of node classification task. default 1e-4

    max_epoch: ``int``
        The max number of epochs in training. default 100

    early_stopping_round: ``int``
        The round of early stop. default 100

    weight_decay: ``float``
        weight decay ratio, default 1e-4

    device: ``torch.device`` or ``str``
        The device where model will be running on.

    init: ``bool``
        If True(False), the model will (not) be initialized.

    feval: (Sequence of) ``Evaluation`` or ``str``
        The evaluation functions, default ``[LogLoss]``
    
    loss: ``str``
        The loss used. Default ``nll_loss``.

    lr_scheduler_type: ``str`` (Optional)
        The lr scheduler type used. Default None.

    """

    def __init__(
        self,
        model: Union[Tuple[BaseEncoderMaintainer, BaseDecoderMaintainer], BaseEncoderMaintainer, BaseAutoModel, str] = None,
        num_features: Optional[int] = None,
        num_classes: Optional[int] = None,
        optimizer: Union[str, Type[torch.optim.Optimizer]] = torch.optim.Adam,
        lr: float = 1e-4,
        max_epoch: int = 100,
        early_stopping_round: int = 100,
        weight_decay: float = 1e-4,
        device: Union[torch.device, str] = "auto",
        init: bool = False,
        feval: Iterable[Type[Evaluation]] =[Logloss],
        loss: Union[Callable, str] = "nll_loss",
        lr_scheduler_type: Optional[str] = None,
        **kwargs
    ):
        if isinstance(model, Tuple):
            encoder, decoder = model
        elif isinstance(model, BaseAutoModel):
            encoder, decoder = model, None
        else:
            encoder, decoder = model, "logsoftmax"

        super().__init__(
            encoder=encoder,
            decoder=decoder,
            num_features=num_features,
            num_classes=num_classes,
            device=device,
            feval=feval,
            loss=loss,
        )

        self.opt_received = optimizer
        if isinstance(optimizer, str):
            if optimizer.lower() == "adam": self.optimizer = torch.optim.Adam
            elif optimizer.lower() == "sgd": self.optimizer = torch.optim.SGD
            else: raise ValueError("Currently not support optimizer {}".format(optimizer))
        elif isinstance(optimizer, type) and issubclass(optimizer, torch.optim.Optimizer):
            self.optimizer = optimizer
        else:
            raise ValueError("Currently not support optimizer {}".format(optimizer))

        self.lr_scheduler_type = lr_scheduler_type
        self.lr = lr
        self.max_epoch = max_epoch
        self.early_stopping_round = early_stopping_round
        self.kwargs = kwargs

        self.weight_decay = weight_decay

        self.early_stopping = EarlyStopping(
            patience=early_stopping_round, verbose=False
        )

        self.valid_result = None
        self.valid_result_prob = None
        self.valid_score = None

        self.pyg_dgl = DependentBackend.get_backend_name()

        self.hyper_parameter_space = [
            {
                "parameterName": "max_epoch",
                "type": "INTEGER",
                "maxValue": 500,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "early_stopping_round",
                "type": "INTEGER",
                "maxValue": 30,
                "minValue": 10,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "lr",
                "type": "DOUBLE",
                "maxValue": 1e-1,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
            {
                "parameterName": "weight_decay",
                "type": "DOUBLE",
                "maxValue": 1e-2,
                "minValue": 1e-4,
                "scalingType": "LOG",
            },
        ]

        self.hyper_parameters = {
            "max_epoch": self.max_epoch,
            "early_stopping_round": self.early_stopping_round,
            "lr": self.lr,
            "weight_decay": self.weight_decay,
        }

        if init is True:
            self.initialize()

    @classmethod
    def get_task_name(cls):
        """
        Derive the task name. (NodeClassification)
        """
        return "NodeClassification"

    def __train_only(self, data, train_mask=None):
        data = data.to(self.device)
        model = self._compose_model()
        if train_mask is None:
            if self.pyg_dgl == 'pyg':
                mask = data.train_mask
            elif self.pyg_dgl == 'dgl':
                mask = data.ndata['train_mask']
        else:
            mask = train_mask
                
        optimizer = self.optimizer(
            model.parameters(),
            lr=self.lr, weight_decay=self.weight_decay
        )
        # scheduler = StepLR(optimizer, step_size=100, gamma=0.1)
        lr_scheduler_type = self.lr_scheduler_type
        if type(lr_scheduler_type) == str and lr_scheduler_type == "steplr":
            scheduler = StepLR(optimizer, step_size=100, gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "multisteplr":
            scheduler = MultiStepLR(optimizer, milestones=[30, 80], gamma=0.1)
        elif type(lr_scheduler_type) == str and lr_scheduler_type == "exponentiallr":
            scheduler = ExponentialLR(optimizer, gamma=0.1)
        elif (
            type(lr_scheduler_type) == str and lr_scheduler_type == "reducelronplateau"
        ):
            scheduler = ReduceLROnPlateau(optimizer, "min")
        else:
            scheduler = None

        for epoch in range(1, self.max_epoch + 1):
            model.train()
            optimizer.zero_grad()
            res = model(data)
            if hasattr(F, self.loss):
                if self.pyg_dgl == 'pyg':
                    loss = getattr(F, self.loss)(res[mask], data.y[mask])
                elif self.pyg_dgl == 'dgl':
                    loss = getattr(F, self.loss)(res[mask], data.ndata['label'][mask])
            else:
                raise TypeError(
                    "PyTorch does not support loss type {}".format(self.loss)
                )

            loss.backward()
            optimizer.step()
            if self.lr_scheduler_type:
                scheduler.step()

            # TODO: move this to autogl.backend.utils
            if self.pyg_dgl == 'pyg' and hasattr(data, "val_mask") and data.val_mask is not None:
                val_mask = data.val_mask
            elif self.pyg_dgl == 'dgl' and data.ndata.get('val_mask', None) is not None:
                val_mask = data.ndata['val_mask']
            else:
                val_mask = None

            if val_mask is not None:
                if type(self.feval) is list:
                    feval = self.feval[0]
                else:
                    feval = self.feval
                val_loss = self.evaluate([data], mask=val_mask, feval=feval)
                if feval.is_higher_better() is True:
                    val_loss = -val_loss

                self.early_stopping(val_loss, model)
                if self.early_stopping.early_stop:
                    LOGGER.debug("Early stopping at %d", epoch)
                    break

        if self.pyg_dgl == "pyg" and hasattr(data, "val_mask") and data.val_mask is not None:
            self.early_stopping.load_checkpoint(model)
        elif self.pyg_dgl == 'dgl' and data.ndata.get('val_mask', None) is not None:
            self.early_stopping.load_checkpoint(model)

    @torch.no_grad()
    def __predict_only(self, data, mask=None):
        if isinstance(mask, str):
            if self.pyg_dgl == 'pyg':
                mask = getattr(data, f'{mask}_mask')
            elif self.pyg_dgl == 'dgl':
                mask = data.ndata[f'{mask}_mask']
        
        model = self._compose_model()
        model.to(self.device)

        data = data.to(self.device)
        model.eval()
        res = model(data)
            
        if mask is None:
            return res
        else:
            return res[mask]

    def train(self, dataset, keep_valid_result=True, train_mask=None):
        """
        Train on the given dataset.

        Parameters
        ----------
        dataset: The node classification dataset used to be trained.

        keep_valid_result: ``bool``
            If True(False), save the validation result after training.

        train_mask: The mask for training data

        Returns
        -------
        self: ``autogl.train.NodeClassificationTrainer``
            A reference of current trainer.

        """
        data = dataset[0]
        self.__train_only(data, train_mask)
        if keep_valid_result:
            if self.pyg_dgl == 'pyg':
                val_mask = data.val_mask
            elif self.pyg_dgl == 'dgl':
                val_mask = data.ndata['val_mask']
            else:
                assert False
            self.valid_result = self.__predict_only(data)[val_mask].max(1)[1]
            self.valid_result_prob = self.__predict_only(data)[val_mask]
            self.valid_score = self.evaluate(
                dataset, mask=val_mask, feval=self.feval
            )
        
    def predict(self, dataset, mask=None):
        """
        Predict on the given dataset using specified mask.

        Parameters
        ----------
        dataset: The node classification dataset used to be predicted.

        mask: ``train``, ``val``, or ``test``.
            The dataset mask.

        Returns
        -------
        The prediction result.
        """
        return self.predict_proba(dataset, mask=mask, in_log_format=True).max(1)[1]

    def predict_proba(self, dataset, mask=None, in_log_format=False):
        """
        Predict the probability on the given dataset using specified mask.

        Parameters
        ----------
        dataset: The node classification dataset used to be predicted.

        mask: ``train``, ``val``, ``test``, or ``Tensor``.
            The dataset mask.

        in_log_format: ``bool``.
            If True(False), the probability will (not) be log format.

        Returns
        -------
        The prediction result.
        """
        data = dataset[0]
        data = data.to(self.device)
        ret = self.__predict_only(data, mask)
        if in_log_format is True:
            return ret
        else:
            return torch.exp(ret)

    def get_valid_predict(self):
        # """Get the valid result."""
        return self.valid_result

    def get_valid_predict_proba(self):
        # """Get the valid result (prediction probability)."""
        return self.valid_result_prob

    def get_valid_score(self, return_major=True):
        """
        The function of getting the valid score.

        Parameters
        ----------
        return_major: ``bool``.
            If True, the return only consists of the major result.
            If False, the return consists of the all results.

        Returns
        -------
        result: The valid score in training stage.
        """
        if isinstance(self.feval, list):
            if return_major:
                return self.valid_score[0], self.feval[0].is_higher_better()
            else:
                return self.valid_score, [f.is_higher_better() for f in self.feval]
        else:
            return self.valid_score, self.feval.is_higher_better()

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(
            {
                "trainer_name": self.__class__.__name__,
                "optimizer": self.optimizer,
                "learning_rate": self.lr,
                "max_epoch": self.max_epoch,
                "early_stopping_round": self.early_stopping_round,
                "encoder": repr(self.encoder),
                "decoder": repr(self.decoder)
            }
        )

    def evaluate(self, dataset, mask=None, feval=None):
        """
        Evaluate on the given dataset.

        Parameters
        ----------
        dataset: The node classification dataset used to be evaluated.

        mask: ``train``, ``val``, or ``test``.
            The dataset mask.

        feval: ``str``.
            The evaluation method used in this function.

        Returns
        -------
        res: The evaluation result on the given dataset.

        """
        data = dataset[0]
        data = data.to(self.device)
        
        if isinstance(mask, str):
            if self.pyg_dgl == 'pyg':
                mask = getattr(data, f'{mask}_mask')
            elif self.pyg_dgl == 'dgl':
                mask = data.ndata[f'{mask}_mask']
        
        if self.pyg_dgl == 'pyg': label = data.y
        elif self.pyg_dgl == 'dgl': label = data.ndata['label']

        if feval is None:
            feval = self.feval
        else:
            feval = get_feval(feval)

        y_pred_prob = self.predict_proba(dataset, mask)
        
        y_true = label[mask] if mask is not None else label

        if not isinstance(feval, list):
            feval = [feval]
            return_signle = True
        else:
            return_signle = False

        res = []
        for f in feval:
            try:
                res.append(f.evaluate(y_pred_prob, y_true))
            except:
                res.append(f.evaluate(y_pred_prob.cpu().numpy(), y_true.cpu().numpy()))
        if return_signle:
            return res[0]
        return res

    def duplicate_from_hyper_parameter(self, hp: dict, model=None, restricted=True):
        """
        The function of duplicating a new instance from the given hyperparameter.

        Parameters
        ----------
        hp: ``dict``.
            The hyperparameter used in the new instance. Should contain 3 keys "trainer", "encoder"
            "decoder", with corresponding hyperparameters as values.

        model:
            Models can be ``str``, ``autogl.module.model.BaseAutoModel``, 
            ``autogl.module.model.encoders.BaseEncoderMaintainer`` or a tuple of (encoder, decoder) 
            if need to specify both encoder and decoder. Encoder can be ``str`` or
            ``autogl.module.model.encoders.BaseEncoderMaintainer``, and decoder can be ``str``
            or ``autogl.module.model.decoders.BaseDecoderMaintainer``.

        restricted: ``bool``.
            If False(True), the hyperparameter should (not) be updated from origin hyperparameter.

        Returns
        -------
        self: ``autogl.train.NodeClassificationTrainer``
            A new instance of trainer.

        """
        if isinstance(model, Tuple):
            encoder, decoder = model
        elif isinstance(model, BaseAutoModel):
            encoder, decoder = model, None
        elif isinstance(model, BaseEncoderMaintainer):
            encoder, decoder = model, self.decoder
        elif model is None:
            encoder, decoder = self.encoder, self.decoder
        else:
            raise TypeError("Cannot parse model with type", type(model))
        
        hp_trainer = hp.get("trainer", {})
        hp_encoder = hp.get("encoder", {})
        hp_decoder = hp.get("decoder", {})
        if not restricted:
            origin_hp = deepcopy(self.hyper_parameters)
            origin_hp.update(hp_trainer)
            hp = origin_hp
        else:
            hp = hp_trainer
        encoder = encoder.from_hyper_parameter(hp_encoder)
        if isinstance(encoder, BaseEncoderMaintainer) and isinstance(decoder, BaseDecoderMaintainer):
            decoder = decoder.from_hyper_parameter_and_encoder(hp_decoder, encoder)

        ret = self.__class__(
            model=(encoder, decoder),
            num_features=self.num_features,
            num_classes=self.num_classes,
            optimizer=self.opt_received,
            lr=hp["lr"],
            max_epoch=hp["max_epoch"],
            early_stopping_round=hp["early_stopping_round"],
            device=self.device,
            weight_decay=hp["weight_decay"],
            feval=self.feval,
            loss=self.loss,
            lr_scheduler_type=self.lr_scheduler_type,
            init=True,
            **self.kwargs
        )

        return ret



from sklearn import preprocessing
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity as cos_sim
import copy
from tqdm import tqdm
from tabulate import tabulate
import time

from .base import BaseFeature, BaseFeatureEngineer
from .selectors import SeGBDT
from . import register_feature

from ...utils import get_logger
import torch

LOGGER = get_logger("Feature")


@register_feature("identity")
class FeIdentity(BaseFeatureEngineer):
    r"""it is a dummy feature engineer , which directly returns identical data"""

    def __init__(self, *args, **kwargs):
        super(FeIdentity, self).__init__(multigraph=True, *args, **kwargs)


@register_feature("onlyconst")
class Onlyconst(BaseFeatureEngineer):
    r"""it is a dummy feature engineer , which directly returns identical data"""

    def __init__(self, *args, **kwargs):
        super(Onlyconst, self).__init__(
            data_t="tensor", multigraph=True, *args, **kwargs
        )

    def _transform(self, data):
        if "x" in data:
            data.x = torch.ones((data.x.shape[0], 1))
        else:
            data.x = torch.ones((torch.unique(data.edge_index).shape[0], 1))
        return data


def op_sum(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.sum(x[nb], axis=0)
    return res


def op_mean(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.mean(x[nb], axis=0)
    return res


def op_max(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.max(x[nb], axis=0)
    return res


def op_min(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.min(x[nb], axis=0)
    return res


def op_prod(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.prod(x[nb], axis=0)
    return res


mms = preprocessing.MinMaxScaler()
ss = preprocessing.StandardScaler()


def scale(x):
    return ss.fit_transform(x)


class Timer:
    def __init__(self, timebudget=None):
        self._timebudget = timebudget
        self._esti_time = 0
        self._g_start = time.time()

    def start(self):
        self._start = time.time()

    def end(self):
        time_use = time.time() - self._start
        self._esti_time = (self._esti_time + time_use) / 2

    def is_timeout(self):
        timebudget = self._timebudget
        if timebudget:
            timebudget = self._timebudget - (time.time() - self._g_start)
            if timebudget < self._esti_time:
                return True
        return False


@register_feature("deepgl")
class AutoFeatureEngineer(BaseFeatureEngineer):
    r"""
    Notes
    -----
    An implementation of auto feature engineering method Deepgl [#]_ ,which iteratively generates features by aggregating neighbour features
    and select a fixed number of  features to automatically add important graph-aware features.
    References
    ----------
    .. [#] Rossi, R. A., Zhou, R., & Ahmed, N. K. (2020).
        Deep Inductive Graph Representation Learning.
        IEEE Transactions on Knowledge and Data Engineering, 32(3), 438–452.
        https://doi.org/10.1109/TKDE.2018.2878247
    Parameters
    ----------
    fixlen : int
        fixed number of features for every epoch. The final number of features added will be
        ``fixlen`` \times ``max_epoch``, 200 \times 5 in default.
    max_epoch : int
        number of epochs in total process.
    timebudget : int
        timebudget(seconds) for the feature engineering process, None for no time budget . Note that
        this time budget is a soft budget ,which is obtained by rough time estimation through previous iterations and
        may finally exceed the actual timebudget
    y_sel_func : Callable
        feature selector function object for selection at each iteration ,lightgbm in default. Note that in original paper,
        connected components of feature graph is used , and you may implement it by yourself if you want.
    verbosity : int
        hide any infomation except error and fatal if ``verbosity`` < 1
    """

    def __init__(
        self,
        fixlen=200,
        max_epoch=5,
        timebudget=None,
        y_sel_func=SeGBDT,
        verbosity=-1,
        *args,
        **kwargs
    ):

        super(AutoFeatureEngineer, self).__init__(multigraph=False, *args, **kwargs)
        self._ops = [op_sum, op_mean, op_max, op_min]
        self._sim = cos_sim
        self._fixlen = fixlen
        self._max_epoch = max_epoch
        self._timebudget = timebudget
        self._y_sel_func = y_sel_func(
            fixlen, verbose_eval=verbosity >= 1, *args, **kwargs
        )
        self._verbosity = verbosity

    def _init(self, data):
        #         self._data=copy.deepcopy(data)
        self._data = data
        self._num_nodes = data.x.shape[0]
        self._x = data.x
        self._edges = data.edge_index
        self._neighbours = [[] for _ in range(self._num_nodes)]
        for u, v in self._edges.T:
            self._neighbours[u].append(v)
        self._neighbours = [np.array(v) for v in self._neighbours]

    def _gen(self, x):
        res = []
        for i, op in enumerate(self._ops):
            res.append(op(x, self._neighbours))
        res = np.concatenate(res, axis=1)
        return res

    def _fit(self, data):
        self._init(data)
        x = self._x.copy()
        gx = x.copy()
        verbs = []
        data = self._data
        max_epoch = self._max_epoch
        timebudget = self._timebudget
        y_sel_func = self._y_sel_func
        soft_timer = Timer(timebudget)
        self._sel = []
        for epoch in tqdm(range(max_epoch), disable=self._verbosity <= 0):
            soft_timer.start()
            verb = [epoch, gx.shape[1]]
            gx = self._gen(gx)
            gx = scale(gx)
            verb.append(gx.shape[1])
            data.x = gx
            #             data = feat_diffuse(data)
            y_sel_func._fit(data)
            self._sel.append(y_sel_func._sel)
            data = y_sel_func._transform(data)
            gx = data.x
            verb.append(gx.shape[1])
            x = np.concatenate([x, gx], axis=1)
            verbs.append(verb)
            soft_timer.end()
            if soft_timer.is_timeout():
                break
        if self._verbosity >= 1:
            LOGGER.info(
                tabulate(verbs, headers="epoch origin after-gen after-sel".split())
            )
        data.x = x
        return data

    def _transform(self, data):
        x = data.x
        gx = x.copy()
        for sel in self._sel:
            gx = self._gen(gx)
            gx = scale(gx)
            gx = gx[:, sel]
            x = np.concatenate([x, gx], axis=1)
        data.x = x
        return data



import torch
import typing as _typing

import torch.nn.functional as F
from dgl.nn.pytorch.conv import SAGEConv
import torch.nn.functional

from . import register_model
from .base import BaseAutoModel, activate_func
from ....utils import get_logger

LOGGER = get_logger("SAGEModel")


class GraphSAGE(torch.nn.Module):

    def __init__(self, args):
        super(GraphSAGE, self).__init__()
        self.args = args
        self.num_layer = int(self.args["num_layers"])

        missing_keys = list(
            set(
                [
                    "features_num",
                    "num_class",
                    "num_layers",
                    "hidden",
                    "dropout",
                    "act",
                    "agg"
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))

        if not self.num_layer == len(self.args["hidden"]) + 1:
            LOGGER.warn("Warning: layer size does not match the length of hidden units")
        
        if self.args["agg"] not in ("gcn", "pool", "mean", "lstm"):
            self.args["agg"] = "gcn"
        
        self.convs = torch.nn.ModuleList()
        self.convs.append(
            SAGEConv(
                self.args["features_num"],
                self.args["hidden"][0],
                aggregator_type=self.args["agg"]
            )
        )
        for i in range(self.num_layer - 2):
            self.convs.append(
                SAGEConv(
                    self.args["hidden"][i] ,
                    self.args["hidden"][i + 1],
                    aggregator_type=self.args["agg"]
                )
            )
            
        self.convs.append(
            SAGEConv(
                self.args["hidden"][-1],
                self.args["num_class"],
                aggregator_type=self.args["agg"]
            )
        )

    def lp_encode(self, data):
        x: torch.Tensor = data.ndata['feat']
        for i in range(len(self.convs) - 2):
            x = self.convs[i](data, x)
            x = activate_func(x, self.args["act"])
        x = self.convs[-2](data, x)
        return x

    def lp_decode(self, z, pos_edge_index, neg_edge_index):
        edge_index = torch.cat([pos_edge_index, neg_edge_index], dim=-1)
        logits = (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)
        return logits

    def lp_decode_all(self, z):
        prob_adj = z @ z.t()
        return (prob_adj > 0).nonzero(as_tuple=False).t()
    
    def forward(self, data):
        try:
            x = data.ndata['feat']
        except:
            print("no x")
            pass

        x = F.dropout(x, p=self.args["dropout"], training=self.training)
        for i in range(self.num_layer):
            x = self.convs[i](data, x)
            if i != self.num_layer - 1:
                x = activate_func(x, self.args["act"])
                x = F.dropout(x, p=self.args["dropout"], training=self.training)

        return F.log_softmax(x, dim=1)



@register_model("sage-model")
class AutoSAGE(BaseAutoModel):
    r"""
    AutoSAGE. The model used in this automodel is GraphSAGE, i.e., the GraphSAGE from the `"Inductive Representation Learning on
    Large Graphs" <https://arxiv.org/abs/1706.02216>`_ paper. The layer is

    .. math::

        \mathbf{x}^{\prime}_i = \mathbf{W}_1 \mathbf{x}_i + \mathbf{W_2} \cdot
        \mathrm{mean}_{j \in \mathcal{N(i)}} \mathbf{x}_j

    Parameters
    ----------
    num_features: `int`.
        The dimension of features.

    num_classes: `int`.
        The number of classes.

    device: `torch.device` or `str`
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.

    """

    def __init__(
        self, num_features=None, num_classes=None, device=None, **args
    ):

        super(AutoSAGE, self).__init__(num_features, num_classes, device, **args)

        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [128, 128, 128],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
            {
                "parameterName": "agg",
                "type": "CATEGORICAL",
                "feasiblePoints": ["mean", "add", "max"],
            },
        ]

        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [64, 32],
            "dropout": 0.5,
            "act": "relu",
            "agg": "mean",
        }

    def _initialize(self):
        self._model = GraphSAGE({
            "features_num": self.input_dimension,
            "num_class": self.output_dimension,
            **self.hyper_parameters
        }).to(self.device)



# codes in this file are reproduced from https://github.com/microsoft/nni with some changes.
import copy
from logging import Logger
from numpy.core.fromnumeric import sort

import torch
import torch.nn as nn
import torch.nn.functional as F

from . import register_nas_algo
from .base import BaseNAS
from ..space import BaseSpace
from ..utils import (
    AverageMeterGroup,
    replace_layer_choice,
    replace_input_choice,
    get_module_order,
    sort_replaced_module,
    PathSamplingLayerChoice,
    PathSamplingInputChoice,
)
from tqdm import tqdm, trange
from ....utils import get_logger

import numpy as np
LOGGER = get_logger("SPOS")

import collections
import dataclasses
import random


@dataclasses.dataclass
class Individual:
    """
    A class that represents an individual.
    Holds two attributes, where ``x`` is the model and ``y`` is the metric (e.g., accuracy).
    """
    x: dict
    y: float


class Evolution:
    """
    Algorithm for regularized evolution (i.e. aging evolution).
    Follows "Algorithm 1" in Real et al. "Regularized Evolution for Image Classifier Architecture Search".
    
    Parameters
    ----------
    optimize_mode : str
        Can be one of "maximize" and "minimize". Default: maximize.
    population_size : int
        The number of individuals to keep in the population. Default: 100.
    cycles : int
        The number of cycles (trials) the algorithm should run for. Default: 20000.
    sample_size : int
        The number of individuals that should participate in each tournament. Default: 25.
    mutation_prob : float
        Probability that mutation happens in each dim. Default: 0.05
    """

    def __init__(self, optimize_mode='maximize', population_size=100, sample_size=25, cycles=20000,
                 mutation_prob=0.05,disable_progress=False):
        assert optimize_mode in ['maximize', 'minimize']
        assert sample_size < population_size
        self.optimize_mode = optimize_mode
        self.population_size = population_size
        self.sample_size = sample_size
        self.cycles = cycles
        self.mutation_prob = mutation_prob
        self.disable_progress= disable_progress
        self._worst = float('-inf') if self.optimize_mode == 'maximize' else float('inf')

        self._success_count = 0
        self._population = collections.deque()
        self._running_models = []
        self._polling_interval = 2.
        self._history = []

    def best_parent(self,sample_size=None):
        """get the config of the best parent 
        """
        samples = [p for p in self._population]  # copy population
        random.shuffle(samples)
        if sample_size is not None:
            samples = list(samples)[:sample_size]
        if self.optimize_mode == 'maximize':
            parent = max(samples, key=lambda sample: sample.y)
        else:
            parent = min(samples, key=lambda sample: sample.y)
        return parent.x

    def _prepare(self):
        self.uniform=UniformSampler(self.nas_modules)
        self.mutation=MutationSampler(self.nas_modules,self.mutation_prob)

    def _get_metric(self,config):
        for name, module in self.nas_modules:
            module.sampled = config[name]
        # todo: this may be computational expensive 
        # model=self.model.parse_model(config,self.device) 
        with torch.no_grad():
            metric, loss = self.estimator.infer(self.model, self.dataset, mask='val')
        return metric[0]

    def search(self, space: BaseSpace,nas_modules,dset,estimator,device):
        self.model = space
        self.dataset = dset 
        self.estimator = estimator
        self.nas_modules = nas_modules
        self.device = device

        self._prepare()
        LOGGER.info('Initializing the first population.')
        with tqdm(range(self.population_size), disable=self.disable_progress) as bar:
            for i in bar:
                config = self.uniform.resample()
                metric=self._get_metric(config)
                individual = Individual(config, metric)
                # LOGGER.debug('Individual created: %s', str(individual))
                self._population.append(individual)
                self._history.append(individual)
                bar.set_postfix(metric=metric,max=max(x.y for x in self._population),min=min(x.y for x in self._population))

        LOGGER.info('Running mutations.')
        with tqdm(range(self.cycles), disable=self.disable_progress) as bar:
            for i in bar:
                parent=self.best_parent(self.sample_size)
                config=self.mutation.resample(parent)
                metric=self._get_metric(config) # todo : add aging factor
                individual = Individual(config, metric)
                LOGGER.debug('Individual created: %s', str(individual))
                self._population.append(individual)
                self._history.append(individual)
                if len(self._population) > self.population_size:
                    self._population.popleft()
                bar.set_postfix(metric=metric,max_h=max(x.y for x in self._history),max=max(x.y for x in self._population),min=min(x.y for x in self._population))
        
        # todo: origin is best in history | or the population may need to be retrained
        self._history.sort(key=lambda x: x.y)
        # best=self.best_parent()
        if self.optimize_mode == 'maximize':
            best=self._history[-1].x
        else:
            best=self._history[0].x
        return best



class MutationSampler:
    """uniform mutator

    Parameters
    ----------
    nas_modules: 
        nas_modules in NAS algorithms , including choices of modules
    mutation_prob: float
        probability of doing mutation in each choice.
    parent : dict
        parent individual's choices
    """
    def __init__(self,nas_modules,mutation_prob):
        selection_range = {}
        for k, v in nas_modules:
            selection_range[k] = len(v)
        self.selection_dict = selection_range
        self.mutation_prob = mutation_prob

    def resample(self, parent):
        search_space=self.selection_dict
        child = {}
        for k, v in parent.items():
            if random.uniform(0, 1) < self.mutation_prob:
                child[k] = np.random.choice(range(search_space[k])) # do not exclude the original operator
            else:
                child[k] = v
        return child

class UniformSampler:
    """Uniform Sampler

    Parameters
    ----------
    nas_modules: 
        nas_modules in NAS algorithms , including choices of modules
    """
    def __init__(self,nas_modules):
        selection_range = {}
        for k, v in nas_modules:
            selection_range[k] = len(v)
        self.selection_dict = selection_range
    def resample(self):
        selection = {}
        for k, v in self.selection_dict.items():
            selection[k] = np.random.choice(range(v))
        return selection

@register_nas_algo("spos")
class Spos(BaseNAS):
    """
    SPOS trainer.

    Parameters
    ----------
    n_warmup : int
        Number of epochs for training super network.
    model_lr : float
        Learning rate for super network.
    model_wd : float
        Weight decay for super network.
    
    Other parameters see Evolution
    """

    def __init__(
        self,
        n_warmup=1000,
        grad_clip=5.0,
        disable_progress=False,
        optimize_mode='maximize', 
        population_size=100, 
        sample_size=25, 
        cycles=20000,
        mutation_prob=0.05,
        device="cuda",
    ):
        super().__init__(device)
        self.model_lr=5e-3
        self.model_wd=5e-4
        self.n_warmup = n_warmup
        self.disable_progress= disable_progress
        self.grad_clip = grad_clip
        self.optimize_mode = optimize_mode
        self.population_size = population_size
        self.sample_size = sample_size
        self.cycles = cycles
        self.mutation_prob = mutation_prob

    def _prepare(self):
        # replace choice
        self.nas_modules = []
        k2o = get_module_order(self.model)
        replace_layer_choice(self.model, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.model, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)
        # to device
        self.model = self.model.to(self.device)
        self.model_optim = torch.optim.Adam(
            self.model.parameters(), lr=self.model_lr, weight_decay=self.model_wd
        )
        # controller
        self.controller=UniformSampler(self.nas_modules)

        # Evolution
        self.evolve = Evolution(
            optimize_mode='maximize', 
            population_size=self.population_size, 
            sample_size=self.sample_size, 
            cycles=self.cycles, 
            mutation_prob=self.mutation_prob,
            disable_progress=self.disable_progress
            )

    def search(self, space: BaseSpace, dset, estimator):
        self.model = space
        self.dataset = dset  
        self.estimator = estimator

        self._prepare()
        self._train() # train using uniform sampling
        self._search() # search using evolutionary algorithm

        selection = self.export()

        # here may sample N , retrain N ,and get best
        print(selection)
        return space.parse_model(selection, self.device)

    def _search(self):
        self.best_config=self.evolve.search(
            self.model,
            self.nas_modules,
            self.dataset,
            self.estimator,
            self.device,
            )


    def _train(self):
        with tqdm(range(self.n_warmup), disable=self.disable_progress) as bar:
            for i in bar:
                acc, l1 = self._train_one_epoch(i)
                with torch.no_grad():
                    val_acc, val_loss = self._infer("val")
                bar.set_postfix(loss=l1, acc=acc, val_acc=val_acc, val_loss=val_loss.item())

    def _train_one_epoch(self, epoch):
        self.model.train()
        self.model_optim.zero_grad()
        self._resample() # uniform sampling
        metric, loss = self._infer(mask="train")
        loss.backward()
        if self.grad_clip > 0:
            nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip)
        self.model_optim.step()
        return metric, loss.item()

    def _resample(self):
        result=self.controller.resample()
        for name, module in self.nas_modules:
            module.sampled = result[name]

    def export(self):
        return self.best_config

    def _infer(self, mask="train"):
        metric, loss = self.estimator.infer(self.model, self.dataset, mask=mask)
        return metric[0], loss



from networkx.algorithms.euler import is_eulerian
from networkx.algorithms.efficiency_measures import global_efficiency
from networkx.algorithms.efficiency_measures import local_efficiency
from networkx.algorithms.distance_regular import is_distance_regular
from networkx.algorithms.components import number_connected_components
from networkx.algorithms.components import is_connected
from networkx.algorithms.cluster import average_clustering
from networkx.algorithms.cluster import transitivity
from networkx.algorithms.clique import graph_number_of_cliques
from networkx.algorithms.clique import graph_clique_number
from networkx.algorithms.bridges import has_bridges
from networkx.algorithms.assortativity import degree_pearson_correlation_coefficient
from networkx.algorithms.assortativity import degree_assortativity_coefficient
from networkx.algorithms.approximation.clustering_coefficient import average_clustering
from networkx.algorithms.approximation.clique import large_clique_size
import netlsd
from .base import BaseGraph
import numpy as np
import torch
from functools import wraps
from .. import register_feature

NX_EXTRACTORS = []


def register_nx(cls):
    NX_EXTRACTORS.append(cls)
    register_feature(cls.__name__)(cls)
    return cls


@register_nx
class NxGraph(BaseGraph):
    def __init__(self, *args, **kwargs):
        super(NxGraph, self).__init__(data_t="nx")
        self._args = args
        self._kwargs = kwargs

    def extract(self, data):
        return data.G.size()

    def _transform(self, data):
        dsc = self.extract(data)
        dsc = torch.FloatTensor([[dsc]])
        data.gf = torch.cat([data.gf, dsc], dim=1)
        return data


def nxfunc(func):
    r"""A decorator for networkx Graph transforms. You may want to use it to quickly wrap a nx Graph feature function object.

    Examples
    --------
    @register_nx
    @nxfunc(large_clique_size)
    class NxLargeCliqueSize(NxGraph):pass

    """

    def decorator_func(cls):
        cls.extract = lambda s, data: func(data.G, *s._args, **s._kwargs)
        return cls

    return decorator_func


@register_nx
@nxfunc(large_clique_size)
class NxLargeCliqueSize(NxGraph):
    pass


@register_nx
@nxfunc(average_clustering)
class NxAverageClusteringApproximate(NxGraph):
    pass


@register_nx
@nxfunc(degree_assortativity_coefficient)
class NxDegreeAssortativityCoefficient(NxGraph):
    pass


@register_nx
@nxfunc(degree_pearson_correlation_coefficient)
class NxDegreePearsonCorrelationCoefficient(NxGraph):
    pass


@register_nx
@nxfunc(has_bridges)
class NxHasBridge(NxGraph):
    pass


@register_nx
@nxfunc(graph_clique_number)
class NxGraphCliqueNumber(NxGraph):
    pass


@register_nx
@nxfunc(graph_number_of_cliques)
class NxGraphNumberOfCliques(NxGraph):
    pass


@register_nx
@nxfunc(transitivity)
class NxTransitivity(NxGraph):
    pass


@register_nx
@nxfunc(average_clustering)
class NxAverageClustering(NxGraph):
    pass


@register_nx
@nxfunc(is_connected)
class NxIsConnected(NxGraph):
    pass


@register_nx
@nxfunc(number_connected_components)
class NxNumberConnectedComponents(NxGraph):
    pass


# from networkx.algorithms.components import is_attracting_component
# @register_nx
# @nxfunc(is_attracting_component)
# class NxIsAttractingComponent(NxGraph):pass

# from networkx.algorithms.components import number_attracting_components
# @register_nx
# @nxfunc(number_attracting_components)
# class NxNumberAttractingComponents(NxGraph):pass

# from networkx.algorithms.connectivity.connectivity import average_node_connectivity
# @register_nx
# @nxfunc(average_node_connectivity)
# class NxAverageNodeConnectivity(NxGraph):pass

# from networkx.algorithms.distance_measures import diameter
# @register_nx
# @nxfunc(diameter)
# class NxDiameter(NxGraph):pass

# from networkx.algorithms.distance_measures import radius
# @register_nx
# @nxfunc(radius)
# class NxRadius(NxGraph):pass


@register_nx
@nxfunc(is_distance_regular)
class NxIsDistanceRegular(NxGraph):
    pass


@register_nx
@nxfunc(local_efficiency)
class NxLocalEfficiency(NxGraph):
    pass


@register_nx
@nxfunc(global_efficiency)
class NxGlobalEfficiency(NxGraph):
    pass


@register_nx
@nxfunc(is_eulerian)
class NxIsEulerian(NxGraph):
    pass


# till algorithms.flows



import numpy as np
import torch
import typing as _typing
from ogb.nodeproppred import NodePropPredDataset
from ogb.linkproppred import LinkPropPredDataset
from ogb.graphproppred import GraphPropPredDataset

from torch_sparse import SparseTensor

from autogl import backend as _backend
from autogl.data import InMemoryStaticGraphSet
from autogl.data.graph import (
    GeneralStaticGraph, GeneralStaticGraphGenerator
)
from ._dataset_registry import DatasetUniversalRegistry
from .utils import index_to_mask


class _OGBDatasetUtil:
    ...


class _OGBNDatasetUtil(_OGBDatasetUtil):
    @classmethod
    def ogbn_data_to_general_static_graph(
            cls, ogbn_data: _typing.Mapping[str, _typing.Union[np.ndarray, int]],
            nodes_label: np.ndarray = ..., nodes_label_key: str = ...,
            train_index: _typing.Optional[np.ndarray] = ...,
            val_index: _typing.Optional[np.ndarray] = ...,
            test_index: _typing.Optional[np.ndarray] = ...,
            nodes_data_key_mapping: _typing.Optional[_typing.Mapping[str, str]] = ...,
            edges_data_key_mapping: _typing.Optional[_typing.Mapping[str, str]] = ...,
            graph_data_key_mapping: _typing.Optional[_typing.Mapping[str, str]] = ...
    ) -> GeneralStaticGraph:
        # TODO
        edge_index = ogbn_data['edge_index']
        num_nodes = ogbn_data['num_nodes']
        edge_feat = ogbn_data['edge_feat']
        if edge_feat is not None:
            edge_feat = torch.tensor(edge_feat)
        edge_index = SparseTensor(row=torch.tensor(edge_index[0]), col=torch.tensor(edge_index[1]), value=edge_feat, sparse_sizes=(num_nodes, num_nodes))
        _, _, value = edge_index.coo()
        if value is not None:
            ogbn_data['edge_feat'] = value.cpu().detach().numpy()
        else:
            ogbn_data['edge_feat'] = edge_feat
        edge_index = edge_index.to_symmetric()
        row, col, _ = edge_index.coo()
        edge_index = np.array([row.cpu().detach().numpy(), col.cpu().detach().numpy()])
        homogeneous_static_graph: GeneralStaticGraph = (
            GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                dict([
                    (target_key, torch.from_numpy(ogbn_data[source_key]))
                    for source_key, target_key in nodes_data_key_mapping.items()
                ]),
                torch.tensor(edge_index),
                dict([
                    (target_key, torch.from_numpy(ogbn_data[source_key]))
                    for source_key, target_key in edges_data_key_mapping.items()
                ]) if isinstance(edges_data_key_mapping, _typing.Mapping) else ...,
                dict([
                    (target_key, torch.from_numpy(ogbn_data[source_key]))
                    for source_key, target_key in graph_data_key_mapping.items()
                ]) if isinstance(graph_data_key_mapping, _typing.Mapping) else ...
            )
        )
        if isinstance(nodes_label, np.ndarray) and isinstance(nodes_label_key, str):
            if ' ' in nodes_label_key:
                raise ValueError("Illegal nodes label key")
            homogeneous_static_graph.nodes.data[nodes_label_key] = (
                torch.from_numpy(nodes_label.squeeze()).squeeze()
            )
        if isinstance(train_index, np.ndarray):
            homogeneous_static_graph.nodes.data['train_mask'] = index_to_mask(
                torch.from_numpy(train_index), ogbn_data['num_nodes']
            )
        if isinstance(val_index, np.ndarray):
            homogeneous_static_graph.nodes.data['val_mask'] = index_to_mask(
                torch.from_numpy(val_index), ogbn_data['num_nodes']
            )
        if isinstance(test_index, np.ndarray):
            homogeneous_static_graph.nodes.data['test_mask'] = index_to_mask(
                torch.from_numpy(test_index), ogbn_data['num_nodes']
            )
        return homogeneous_static_graph

    @classmethod
    def ogbn_dataset_to_general_static_graph(
            cls, ogbn_dataset: NodePropPredDataset,
            nodes_label_key: str,
            nodes_data_key_mapping: _typing.Optional[_typing.Mapping[str, str]] = ...,
            edges_data_key_mapping: _typing.Optional[_typing.Mapping[str, str]] = ...,
            graph_data_key_mapping: _typing.Optional[_typing.Mapping[str, str]] = ...
    ) -> GeneralStaticGraph:
        split_idx = ogbn_dataset.get_idx_split()
        return cls.ogbn_data_to_general_static_graph(
            ogbn_dataset[0][0],
            ogbn_dataset[0][1],
            nodes_label_key,
            split_idx["train"],
            split_idx["valid"],
            split_idx["test"],
            nodes_data_key_mapping,
            edges_data_key_mapping,
            graph_data_key_mapping
        )


@DatasetUniversalRegistry.register_dataset("ogbn-products")
class OGBNProductsDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbn_dataset = NodePropPredDataset("ogbn-products", path)
        if _backend.DependentBackend.is_dgl():
            super(OGBNProductsDataset, self).__init__([
                _OGBNDatasetUtil.ogbn_dataset_to_general_static_graph(
                    ogbn_dataset, "label",
                    {"node_feat": "feat"},
                    {"edge_feat": "edge_feat"}
                )
            ])
        elif _backend.DependentBackend.is_pyg():
            super(OGBNProductsDataset, self).__init__([
                _OGBNDatasetUtil.ogbn_dataset_to_general_static_graph(
                    ogbn_dataset, "y",
                    {"node_feat": "x"}
                )
            ])


@DatasetUniversalRegistry.register_dataset("ogbn-proteins")
class OGBNProteinsDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbn_dataset = NodePropPredDataset("ogbn-proteins", path)
        if _backend.DependentBackend.is_dgl():
            super(OGBNProteinsDataset, self).__init__([
                _OGBNDatasetUtil.ogbn_dataset_to_general_static_graph(
                    ogbn_dataset, "label",
                    {"node_species": "species"},
                    {"edge_feat": "edge_feat"}
                )
            ])
        elif _backend.DependentBackend.is_pyg():
            super(OGBNProteinsDataset, self).__init__([
                _OGBNDatasetUtil.ogbn_dataset_to_general_static_graph(
                    ogbn_dataset, "y",
                    {"node_species": "species"},
                    {"edge_feat": "edge_feat"}
                )
            ])


@DatasetUniversalRegistry.register_dataset("ogbn-arxiv")
class OGBNArxivDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbn_dataset = NodePropPredDataset("ogbn-arxiv", path)
        if _backend.DependentBackend.is_dgl():
            super(OGBNArxivDataset, self).__init__([
                _OGBNDatasetUtil.ogbn_dataset_to_general_static_graph(
                    ogbn_dataset, "label",
                    {
                        "node_feat": "feat",
                        "node_year": "year"
                    }
                )
            ])
        elif _backend.DependentBackend.is_pyg():
            super(OGBNArxivDataset, self).__init__([
                _OGBNDatasetUtil.ogbn_dataset_to_general_static_graph(
                    ogbn_dataset, "y",
                    {
                        "node_feat": "x",
                        "node_year": "year"
                    }
                )
            ])


@DatasetUniversalRegistry.register_dataset("ogbn-papers100M")
class OGBNPapers100MDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbn_dataset = NodePropPredDataset("ogbn-papers100M", path)
        if _backend.DependentBackend.is_dgl():
            super(OGBNPapers100MDataset, self).__init__([
                _OGBNDatasetUtil.ogbn_dataset_to_general_static_graph(
                    ogbn_dataset, "label",
                    {
                        "node_feat": "feat",
                        "node_year": "year"
                    }
                )
            ])
        elif _backend.DependentBackend.is_pyg():
            super(OGBNPapers100MDataset, self).__init__([
                _OGBNDatasetUtil.ogbn_dataset_to_general_static_graph(
                    ogbn_dataset, "y",
                    {
                        "node_feat": "x",
                        "node_year": "year"
                    }
                )
            ])


# todo: currently homogeneous dataset `ogbn-mag` NOT supported


class _OGBLDatasetUtil(_OGBDatasetUtil):
    @classmethod
    def ogbl_data_to_general_static_graph(
            cls, ogbl_data: _typing.Mapping[str, _typing.Union[np.ndarray, int]],
            heterogeneous_edges: _typing.Mapping[
                _typing.Tuple[str, str, str],
                _typing.Union[
                    torch.Tensor,
                    _typing.Tuple[torch.Tensor, _typing.Optional[_typing.Mapping[str, torch.Tensor]]]
                ]
            ] = ...,
            nodes_data_key_mapping: _typing.Optional[_typing.Mapping[str, str]] = ...,
            graph_data_key_mapping: _typing.Optional[_typing.Mapping[str, str]] = ...
    ) -> GeneralStaticGraph:
        return GeneralStaticGraphGenerator.create_heterogeneous_static_graph(
            {
                '': dict([
                    (target_data_key, torch.from_numpy(ogbl_data[source_data_key]).squeeze())
                    for source_data_key, target_data_key in nodes_data_key_mapping.items()
                ])
            },
            heterogeneous_edges,
            dict([
                (target_data_key, torch.from_numpy(ogbl_data[source_data_key]).squeeze())
                for source_data_key, target_data_key in graph_data_key_mapping.items()
            ]) if isinstance(graph_data_key_mapping, _typing.Mapping) else ...
        )


@DatasetUniversalRegistry.register_dataset("ogbl-ppa")
class OGBLPPADataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbl_dataset = LinkPropPredDataset("ogbl-ppa", path)
        edge_split = ogbl_dataset.get_edge_split()
        super(OGBLPPADataset, self).__init__([
            _OGBLDatasetUtil.ogbl_data_to_general_static_graph(
                ogbl_dataset[0], {
                    ('', '', ''): torch.from_numpy(ogbl_dataset[0]['edge_index']),
                    ('', 'train_pos_edge', ''): torch.from_numpy(edge_split['train']['edge']),
                    ('', 'val_pos_edge', ''): torch.from_numpy(edge_split['valid']['edge']),
                    ('', 'val_neg_edge', ''): torch.from_numpy(edge_split['valid']['edge_neg']),
                    ('', 'test_pos_edge', ''): torch.from_numpy(edge_split['test']['edge']),
                    ('', 'test_neg_edge', ''): torch.from_numpy(edge_split['test']['edge_neg'])
                },
                {'node_feat': 'feat'} if _backend.DependentBackend.is_dgl() else {'node_feat': 'x'}
            )
        ])


@DatasetUniversalRegistry.register_dataset("ogbl-collab")
class OGBLCOLLABDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbl_dataset = LinkPropPredDataset("ogbl-collab", path)
        edge_split = ogbl_dataset.get_edge_split()
        super(OGBLCOLLABDataset, self).__init__([
            _OGBLDatasetUtil.ogbl_data_to_general_static_graph(
                ogbl_dataset[0], {
                    ('', '', ''): torch.from_numpy(ogbl_dataset[0]['edge_index']),
                    ('', 'train_pos_edge', ''): (
                        torch.from_numpy(edge_split['train']['edge']),
                        {
                            'weight': torch.from_numpy(edge_split['train']['weight']),
                            'year': torch.from_numpy(edge_split['train']['year'])
                        }
                    ),
                    ('', 'val_pos_edge', ''): (
                        torch.from_numpy(edge_split['valid']['edge']),
                        {
                            'weight': torch.from_numpy(edge_split['valid']['weight']),
                            'year': torch.from_numpy(edge_split['valid']['year'])
                        }
                    ),
                    ('', 'val_neg_edge', ''): torch.from_numpy(edge_split['valid']['edge_neg']),
                    ('', 'test_pos_edge', ''): (
                        torch.from_numpy(edge_split['test']['edge']),
                        {
                            'weight': torch.from_numpy(edge_split['test']['weight']),
                            'year': torch.from_numpy(edge_split['test']['year'])
                        }
                    ),
                    ('', 'test_neg_edge', ''): torch.from_numpy(edge_split['test']['edge_neg'])
                },
                {'node_feat': 'feat'} if _backend.DependentBackend.is_dgl() else {'node_feat': 'x'}
            )
        ])


@DatasetUniversalRegistry.register_dataset("ogbl-ddi")
class OGBLDDIDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbl_dataset = LinkPropPredDataset("ogbl-ddi", path)
        edge_split = ogbl_dataset.get_edge_split()
        super(OGBLDDIDataset, self).__init__([
            GeneralStaticGraphGenerator.create_heterogeneous_static_graph(
                {'': {'_NID': torch.arange(ogbl_dataset[0]['num_nodes'])}},
                {
                    ('', '', ''): torch.from_numpy(ogbl_dataset[0]['edge_index']),
                    ('', 'train_pos_edge', ''): torch.from_numpy(edge_split['train']['edge']),
                    ('', 'val_pos_edge', ''): torch.from_numpy(edge_split['valid']['edge']),
                    ('', 'val_neg_edge', ''): torch.from_numpy(edge_split['valid']['edge_neg']),
                    ('', 'test_pos_edge', ''): torch.from_numpy(edge_split['test']['edge']),
                    ('', 'test_neg_edge', ''): torch.from_numpy(edge_split['test']['edge_neg'])
                }
            )
        ])


@DatasetUniversalRegistry.register_dataset("ogbl-citation")
@DatasetUniversalRegistry.register_dataset("ogbl-citation2")
class OGBLCitation2Dataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbl_dataset = LinkPropPredDataset("ogbl-citation2", path)
        edge_split = ogbl_dataset.get_edge_split()
        super(OGBLCitation2Dataset, self).__init__([
            _OGBLDatasetUtil.ogbl_data_to_general_static_graph(
                ogbl_dataset[0],
                {
                    ('', '', ''): torch.from_numpy(ogbl_dataset[0]['edge_index']),
                    ('', 'train_pos_edge', ''): torch.from_numpy(edge_split['train']['edge']),
                    ('', 'val_pos_edge', ''): torch.from_numpy(edge_split['valid']['edge']),
                    ('', 'val_neg_edge', ''): torch.from_numpy(edge_split['valid']['edge_neg']),
                    ('', 'test_pos_edge', ''): torch.from_numpy(edge_split['test']['edge']),
                    ('', 'test_neg_edge', ''): torch.from_numpy(edge_split['test']['edge_neg'])
                },
                (
                    {'node_feat': 'feat', 'node_year': 'year'}
                    if _backend.DependentBackend.is_dgl()
                    else {'node_feat': 'x', 'node_year': 'year'}
                )
            )
        ])


# todo: currently homogeneous dataset `ogbl-wikikg2` and `ogbl-biokg` NOT supported


class _OGBGDatasetUtil:
    ...


@DatasetUniversalRegistry.register_dataset("ogbg-molhiv")
class OGBGMOLHIVDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbl_dataset = GraphPropPredDataset("ogbg-molhiv", path)
        idx_split: _typing.Mapping[str, np.ndarray] = ogbl_dataset.get_idx_split()
        train_index: _typing.Any = idx_split['train'].tolist()
        test_index: _typing.Any = idx_split['test'].tolist()
        val_index: _typing.Any = idx_split['valid'].tolist()
        super(OGBGMOLHIVDataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    (
                        {"feat": torch.from_numpy(data['node_feat'])}
                        if _backend.DependentBackend.is_dgl()
                        else {"x": torch.from_numpy(data['node_feat'])}
                    ),
                    torch.from_numpy(data['edge_index']),
                    {'edge_feat': torch.from_numpy(data['edge_feat'])},
                    (
                        {'label': torch.from_numpy(label)}
                        if _backend.DependentBackend.is_dgl()
                        else {'y': torch.from_numpy(label)}
                    )
                ) for data, label in ogbl_dataset
            ],
            train_index, val_index, test_index
        )


@DatasetUniversalRegistry.register_dataset("ogbg-molpcba")
class OGBGMOLPCBADataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbl_dataset = GraphPropPredDataset("ogbg-molhiv", path)
        idx_split: _typing.Mapping[str, np.ndarray] = ogbl_dataset.get_idx_split()
        train_index: _typing.Any = idx_split['train'].tolist()
        test_index: _typing.Any = idx_split['test'].tolist()
        val_index: _typing.Any = idx_split['valid'].tolist()
        super(OGBGMOLPCBADataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    (
                        {"feat": torch.from_numpy(data['node_feat'])}
                        if _backend.DependentBackend.is_dgl()
                        else {"x": torch.from_numpy(data['node_feat'])}
                    ),
                    torch.from_numpy(data['edge_index']),
                    {'edge_feat': torch.from_numpy(data['edge_feat'])},
                    (
                        {'label': torch.from_numpy(label)}
                        if _backend.DependentBackend.is_dgl()
                        else {'y': torch.from_numpy(label)}
                    )
                ) for data, label in ogbl_dataset
            ],
            train_index, val_index, test_index
        )


@DatasetUniversalRegistry.register_dataset("ogbg-ppa")
class OGBGPPADataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbl_dataset = GraphPropPredDataset("ogbg-molhiv", path)
        idx_split: _typing.Mapping[str, np.ndarray] = ogbl_dataset.get_idx_split()
        train_index: _typing.Any = idx_split['train'].tolist()
        test_index: _typing.Any = idx_split['test'].tolist()
        val_index: _typing.Any = idx_split['valid'].tolist()
        super(OGBGPPADataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    {'_NID': torch.arange(data['num_nodes'])},
                    torch.from_numpy(data['edge_index']),
                    {'edge_feat': torch.from_numpy(data['edge_feat'])},
                    (
                        {'label': torch.from_numpy(label)}
                        if _backend.DependentBackend.is_dgl()
                        else {'y': torch.from_numpy(label)}
                    )
                ) for data, label in ogbl_dataset
            ],
            train_index, val_index, test_index
        )


@DatasetUniversalRegistry.register_dataset("ogbg-code")
@DatasetUniversalRegistry.register_dataset("ogbg-code2")
class OGBGCode2Dataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        ogbl_dataset = GraphPropPredDataset("ogbg-molhiv", path)
        idx_split: _typing.Mapping[str, np.ndarray] = ogbl_dataset.get_idx_split()
        train_index: _typing.Any = idx_split['train'].tolist()
        test_index: _typing.Any = idx_split['test'].tolist()
        val_index: _typing.Any = idx_split['valid'].tolist()
        super(OGBGCode2Dataset, self).__init__(
            [
                GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                    (
                        {
                            "feat": torch.from_numpy(data['node_feat']),
                            "node_is_attributed": torch.from_numpy(data["node_is_attributed"]),
                            "node_dfs_order": torch.from_numpy(data["node_dfs_order"]),
                            "node_depth": torch.from_numpy(data["node_depth"])
                        }
                        if _backend.DependentBackend.is_dgl()
                        else
                        {
                            "x": torch.from_numpy(data['node_feat']),
                            "node_is_attributed": torch.from_numpy(data["node_is_attributed"]),
                            "node_dfs_order": torch.from_numpy(data["node_dfs_order"]),
                            "node_depth": torch.from_numpy(data["node_depth"])
                        }
                    ),
                    torch.from_numpy(data['edge_index'])
                ) for data, label in ogbl_dataset
            ],
            train_index, val_index, test_index
        )



import torch
import typing as _typ
import torch.nn as nn
import torch.nn.functional as F

from . import register_nas_space
from .base import BaseSpace, map_nn
from ...model import BaseAutoModel
from .operation import act_map
from ..utils import count_parameters, measure_latency

from ..backend import *

from operator import *
from .operation import *

@register_nas_space("graphnasmacro")
class GraphNasMacroNodeClassificationSpace(BaseSpace):
    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.6,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = None,
        search_act_con=False,
    ):
        super().__init__()
        self.layer_number = layer_number
        self.hidden_dim = hidden_dim
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.ops = ops
        self.dropout = dropout
        self.search_act_con = search_act_con

    def instantiate(
        self,
        hidden_dim: _typ.Optional[int] = None,
        layer_number: _typ.Optional[int] = None,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = None,
        dropout=None,
    ):
        super().instantiate()
        self.hidden_dim = hidden_dim or self.hidden_dim
        self.layer_number = layer_number or self.layer_number
        self.input_dim = input_dim or self.input_dim
        self.output_dim = output_dim or self.output_dim
        self.ops = ops or self.ops
        self.dropout = dropout or self.dropout

        num_feat = self.input_dim
        num_label = self.output_dim

        layer_nums = self.layer_number
        state_num = 5

        # build hidden layer
        for i in range(layer_nums):
            # extract layer information
            setattr(
                self,
                f"attention_{i}",
                self.setLayerChoice(
                    i * state_num + 0,
                    map_nn(
                        [
                            "gat",
                            "gcn",
                            "cos",
                            "const",
                            "gat_sym",
                            "linear",
                            "generalized_linear",
                        ]
                    ),
                    key=f"attention_{i}",
                ),
            )
            setattr(
                self,
                f"aggregator_{i}",
                self.setLayerChoice(
                    i * state_num + 1,
                    map_nn(
                        [
                            "sum",
                            "mean",
                            "max",
                            "mlp",
                        ]
                    ),
                    key=f"aggregator_{i}",
                ),
            )
            setattr(
                self,
                f"act_{i}",
                self.setLayerChoice(
                    i * state_num + 0,
                    map_nn(
                        [
                            "sigmoid",
                            "tanh",
                            "relu",
                            "linear",
                            "softplus",
                            "leaky_relu",
                            "relu6",
                            "elu",
                        ]
                    ),
                    key=f"act_{i}",
                ),
            )
            setattr(
                self,
                f"head_{i}",
                self.setLayerChoice(
                    i * state_num + 0, map_nn([1, 2, 4, 6, 8, 16]), key=f"head_{i}"
                ),
            )
            if i < layer_nums - 1:
                setattr(
                    self,
                    f"out_channels_{i}",
                    self.setLayerChoice(
                        i * state_num + 0,
                        map_nn([4, 8, 16, 32, 64, 128, 256]),
                        key=f"out_channels_{i}",
                    ),
                )

    def parse_model(self, selection, device) -> BaseAutoModel:
        sel_list = []
        for i in range(self.layer_number):
            sel_list.append(
                [
                    "gat",
                    "gcn",
                    "cos",
                    "const",
                    "gat_sym",
                    "linear",
                    "generalized_linear",
                ][selection[f"attention_{i}"]]
            )
            sel_list.append(
                [
                    "sum",
                    "mean",
                    "max",
                    "mlp",
                ][selection[f"aggregator_{i}"]]
            )
            sel_list.append(
                [
                    "sigmoid",
                    "tanh",
                    "relu",
                    "linear",
                    "softplus",
                    "leaky_relu",
                    "relu6",
                    "elu",
                ][selection[f"act_{i}"]]
            )
            sel_list.append([1, 2, 4, 6, 8, 16][selection[f"head_{i}"]])
            if i < self.layer_number - 1:
                sel_list.append(
                    [4, 8, 16, 32, 64, 128, 256][selection[f"out_channels_{i}"]]
                )
        sel_list.append(self.output_dim)
        # sel_list = ['const', 'sum', 'relu6', 2, 128, 'gat', 'sum', 'linear', 2, 7]
        model = GraphNet(
            sel_list,
            self.input_dim,
            self.output_dim,
            self.dropout,
            multi_label=False,
            batch_normal=False,
            layers=self.layer_number,
        ).wrap()
        return model


class GraphNet(BaseSpace):
    def __init__(
        self,
        actions,
        num_feat,
        num_label,
        drop_out=0.6,
        multi_label=False,
        batch_normal=True,
        state_num=5,
        residual=False,
        layers=2,
    ):
        self.residual = residual
        self.batch_normal = batch_normal
        self.layer_nums = layers
        self.multi_label = multi_label
        self.num_feat = num_feat
        self.num_label = num_label
        self.input_dim = num_feat
        self.output_dim = num_label
        self.dropout = drop_out

        super().__init__()
        self.build_model(
            actions, batch_normal, drop_out, num_feat, num_label, state_num
        )

    def build_model(
        self, actions, batch_normal, drop_out, num_feat, num_label, state_num
    ):
        if self.residual:
            self.fcs = torch.nn.ModuleList()
        if self.batch_normal:
            self.bns = torch.nn.ModuleList()
        self.layers = torch.nn.ModuleList()
        self.acts = []
        self.gates = torch.nn.ModuleList()
        self.build_hidden_layers(
            actions,
            batch_normal,
            drop_out,
            self.layer_nums,
            num_feat,
            num_label,
            state_num,
        )

    def build_hidden_layers(
        self,
        actions,
        batch_normal,
        drop_out,
        layer_nums,
        num_feat,
        num_label,
        state_num=6,
    ):

        # build hidden layer
        for i in range(layer_nums):

            if i == 0:
                in_channels = num_feat
            else:
                in_channels = out_channels * head_num

            # extract layer information
            attention_type = actions[i * state_num + 0]
            aggregator_type = actions[i * state_num + 1]
            act = actions[i * state_num + 2]
            head_num = actions[i * state_num + 3]
            out_channels = actions[i * state_num + 4]
            concat = True
            if i == layer_nums - 1:
                concat = False
            if self.batch_normal:
                self.bns.append(torch.nn.BatchNorm1d(in_channels, momentum=0.5))
            self.layers.append(
                GeoLayer(
                    in_channels,
                    out_channels,
                    head_num,
                    concat,
                    dropout=self.dropout,
                    att_type=attention_type,
                    agg_type=aggregator_type,
                )
            )
            self.acts.append(act_map(act))
            if self.residual:
                if concat:
                    self.fcs.append(
                        torch.nn.Linear(in_channels, out_channels * head_num)
                    )
                else:
                    self.fcs.append(torch.nn.Linear(in_channels, out_channels))

    def forward(self, data):
        output=bk_feat(data)
        # output, edge_index_all = data.x, data.edge_index  # x [2708,1433] ,[2, 10556]
        if self.residual:
            for i, (act, layer, fc) in enumerate(zip(self.acts, self.layers, self.fcs)):
                output = F.dropout(output, p=self.dropout, training=self.training)
                if self.batch_normal:
                    output = self.bns[i](output)

                # output = act(layer(output, edge_index_all) + fc(output))
                output = act(bk_gconv(layer,data,output) + fc(output))
        else:
            for i, (act, layer) in enumerate(zip(self.acts, self.layers)):
                output = F.dropout(output, p=self.dropout, training=self.training)
                if self.batch_normal:
                    output = self.bns[i](output)
                # output = act(layer(output, edge_index_all))
                output = act(bk_gconv(layer,data,output))

        if not self.multi_label:
            output = F.log_softmax(output, dim=1)
        return output

    def __repr__(self):
        result_lines = ""
        for each in self.layers:
            result_lines += str(each)
        return result_lines

    @staticmethod
    def merge_param(old_param, new_param, update_all):
        for key in new_param:
            if update_all or key not in old_param:
                old_param[key] = new_param[key]
        return old_param

    def get_param_dict(self, old_param=None, update_all=True):
        if old_param is None:
            result = {}
        else:
            result = old_param
        for i in range(self.layer_nums):
            key = "layer_%d" % i
            new_param = self.layers[i].get_param_dict()
            if key in result:
                new_param = self.merge_param(result[key], new_param, update_all)
                result[key] = new_param
            else:
                result[key] = new_param
        if self.residual:
            for i, fc in enumerate(self.fcs):
                key = f"layer_{i}_fc_{fc.weight.size(0)}_{fc.weight.size(1)}"
                result[key] = self.fcs[i]
        if self.batch_normal:
            for i, bn in enumerate(self.bns):
                key = f"layer_{i}_fc_{bn.weight.size(0)}"
                result[key] = self.bns[i]
        return result

    def load_param(self, param):
        if param is None:
            return

        for i in range(self.layer_nums):
            self.layers[i].load_param(param["layer_%d" % i])

        if self.residual:
            for i, fc in enumerate(self.fcs):
                key = f"layer_{i}_fc_{fc.weight.size(0)}_{fc.weight.size(1)}"
                if key in param:
                    self.fcs[i] = param[key]
        if self.batch_normal:
            for i, bn in enumerate(self.bns):
                key = f"layer_{i}_fc_{bn.weight.size(0)}"
                if key in param:
                    self.bns[i] = param[key]



import numpy as np
from .. import register_feature
from ..base import BaseFeature


class BaseGenerator(BaseFeature):
    def __init__(self, data_t="np", multigraph=True, **kwargs):
        super(BaseGenerator, self).__init__(
            data_t=data_t, multigraph=multigraph, **kwargs
        )


@register_feature("onehot")
class GeOnehot(BaseGenerator):
    def _transform(self, data):
        fe = np.eye(data.x.shape[0])
        data.x = np.concatenate([data.x, fe], axis=1)
        return data



from autogl.module.nas.space.operation import gnn_map
import typing as _typ
import torch

import torch.nn.functional as F

from . import register_nas_space
from .base import apply_fixed_architecture
from .base import BaseSpace
from ...model import BaseAutoModel
from ....utils import get_logger

from ...model import AutoGCN
from ..backend import *
from ..utils import count_parameters, measure_latency

@register_nas_space("singlepath")
class SinglePathNodeClassificationSpace(BaseSpace):
    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.2,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = ['gcn', "gat_8"],
        # ops: _typ.Tuple = [ "gat_8"],

    ):
        super().__init__()
        self.layer_number = layer_number
        self.hidden_dim = hidden_dim
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.ops = ops
        self.dropout = dropout

    def instantiate(
        self,
        hidden_dim: _typ.Optional[int] = None,
        layer_number: _typ.Optional[int] = None,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = None,
        dropout=None,
    ):
        super().instantiate()
        self.hidden_dim = hidden_dim or self.hidden_dim
        self.layer_number = layer_number or self.layer_number
        self.input_dim = input_dim or self.input_dim
        self.output_dim = output_dim or self.output_dim
        self.ops = ops or self.ops
        self.dropout = dropout or self.dropout
        for layer in range(self.layer_number):
            setattr(
                self,
                f"op_{layer}",
                self.setLayerChoice(
                    layer,
                    [
                        op(
                            self.input_dim if layer == 0 else self.hidden_dim,
                            self.output_dim
                            if layer == self.layer_number - 1
                            else self.hidden_dim,
                        )
                        if isinstance(op, type)
                        else gnn_map(
                            op,
                            self.input_dim if layer == 0 else self.hidden_dim,
                            self.output_dim
                            if layer == self.layer_number - 1
                            else self.hidden_dim,
                        )
                        for op in self.ops
                    ],
                ),
            )
        self._initialized = True

    def forward(self, data):
        x= bk_feat(data)
        for layer in range(self.layer_number):
            op= getattr(self, f"op_{layer}")
            x = bk_gconv(op,data,x)
            if layer != self.layer_number - 1:
                x = F.leaky_relu(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        return F.log_softmax(x, dim=1)

    def parse_model(self, selection, device) -> BaseAutoModel:
        # return AutoGCN(self.input_dim, self.output_dim, device)
        return self.wrap().fix(selection)
